#!/bin/env python3
import os
import subprocess
import argparse
import json
from elftools.elf.elffile import ELFFile


#meson changed behaviour from 0.49 to 0.50 so we need this:
def meson_fetch_filename(filename_object):
  if isinstance(filename_object, str):
    return filename_object
  else:
    return filename_object[0]


def needed_libs(filename):
  print('Processing file:', filename)
  result = []
  with open(filename, 'rb') as f:
    elffile = ELFFile(f)
    for section in elffile.iter_sections():
      if section.name.startswith('.dynamic'):
        for tag in section.iter_tags():
          if tag.entry.d_tag == 'DT_NEEDED':
            result.append(getattr(tag, tag.entry.d_tag[3:].lower()))
  return result


parser = argparse.ArgumentParser(description='Check that when build with efl-one that no module nor efl-one lib does drag in libeina or the likes')
parser.add_argument('builddir', metavar='build', help='the path where to find the meson build directory')

G = parser.parse_args()

#Run meson to fetch all examples
meson_introspect = subprocess.Popen(["meson", "introspect", G.builddir, "--targets"],
      stdout = subprocess.PIPE,
      stderr = subprocess.PIPE,
)
meson_introspect.poll()
build_targets = json.loads(meson_introspect.stdout.read())
build_modules = [meson_fetch_filename(b["filename"]) for b in build_targets if "modules" in meson_fetch_filename(b["filename"]) and meson_fetch_filename(b["filename"]).endswith('.so')]

for build_modules in build_modules:
  libs = needed_libs(build_modules)
  lib_es = [lib for lib in libs if lib.startswith("libe") and lib != "libefl-one.so.1"]
  if len(lib_es) != 0:
    print("Error, {} requies lib {}".format(build_modules, lib_es[0]))
    exit(-1)

print("Nothing wrong found!")




#!/usr/bin/env python3
# encoding: utf-8
"""
Pyolian: ctypes bindings for the eolian library.

This file (along with eolian_lib.py) implement full pythonic eolian bindings.

Those bindings are designed to work at runtime without any prepare/compilation
steps required. The only requirement is that we can find the eolian.so/dll
somwhere built in the source tree, look at eolian_lib.py to see the searchs
performed.

You can directly use the python API provided here if you need direct access
to eolian, or we suggest to look at the template-based generator.py if you just
need to generate some sort of text files out of the eolian database.

To use this library from outside this directory, you need to hack sys.path in
a way that this folder will be available on PYTHON_PATH, fe:

  pyolian_path = os.path.join(EFL_ROOT_PATH, 'src', 'scripts')
  sys.path.insert(0, pyolian_path)
  from pyolian import eolian
  from pyolian.generator import Template

"""
import os
from enum import IntEnum
import atexit
from ctypes import cast, byref, c_char_p, c_void_p, c_int
import ctypes

try:
    from .eolian_lib import lib
except ImportError:
    from eolian_lib import lib

_already_halted = False


# This is the same as the EOLIAN_FILE_FORMAT_VERSION macro
file_format_version = lib.eolian_file_format_version_get()


#  Eolian Enums  ##############################################################
class Eolian_Object_Type(IntEnum):
    UNKNOWN = 0
    CLASS = 1
    TYPEDECL = 2
    STRUCT_FIELD = 3
    ENUM_FIELD = 4
    TYPE = 5
    CONSTANT = 6
    EXPRESSION = 7
    FUNCTION = 8
    FUNCTION_PARAMETER = 9
    EVENT = 10
    PART = 11
    IMPLEMENT = 12
    CONSTRUCTOR = 13
    DOCUMENTATION = 14
    ERROR = 15


class Eolian_Function_Type(IntEnum):
    UNRESOLVED = 0
    PROPERTY = 1
    PROP_SET = 2
    PROP_GET = 3
    METHOD = 4
    FUNCTION_POINTER = 5


class Eolian_Parameter_Direction(IntEnum):
    UNKNOWN = 0
    IN = 1
    OUT = 2
    INOUT = 3


class Eolian_Class_Type(IntEnum):
    UNKNOWN_TYPE = 0
    REGULAR = 1
    ABSTRACT = 2
    MIXIN = 3
    INTERFACE = 4


class Eolian_Object_Scope(IntEnum):
    UNKNOWN = 0
    PUBLIC = 1
    PRIVATE = 2
    PROTECTED = 3


class Eolian_Typedecl_Type(IntEnum):
    UNKNOWN = 0
    STRUCT = 1
    STRUCT_OPAQUE = 2
    ENUM = 3
    ALIAS = 4
    FUNCTION_POINTER = 5


class Eolian_Type_Type(IntEnum):
    UNKNOWN_TYPE = 0
    VOID = 1
    REGULAR = 2
    CLASS = 3
    UNDEFINED = 4


class Eolian_Type_Builtin_Type(IntEnum):
    INVALID = 0
    BYTE = 1
    UBYTE = 2
    CHAR = 3
    SHORT = 4
    USHORT = 5
    INT = 6
    UINT = 7
    LONG = 8
    ULONG = 9
    LLONG = 10
    ULLONG = 11

    INT8 = 12
    UINT8 = 13
    INT16 = 14
    UINT16 = 15
    INT32 = 16
    UINT32 = 17
    INT64 = 18
    UINT64 = 19
    INT128 = 20
    UINT128 = 21

    SIZE = 22
    SSIZE = 23
    INTPTR = 24
    UINTPTR = 25
    PTRDIFF = 26

    TIME = 27

    FLOAT = 28
    DOUBLE = 29

    BOOL = 30

    SLICE = 31
    RW_SLICE = 32

    VOID = 33

    ACCESSOR = 34
    ARRAY = 35
    FUTURE = 36
    ITERATOR = 37
    LIST = 38

    ANY_VALUE = 39
    ANY_VALUE_REF = 40
    BINBUF = 41
    EVENT = 42

    MSTRING = 43
    STRING = 44
    STRINGSHARE = 45
    STRBUF = 46

    HASH = 47
    VOID_PTR = 48


class Eolian_Expression_Type(IntEnum):
    UNKNOWN = 0
    INT = 1
    UINT = 2
    LONG = 3
    ULONG = 4
    LLONG = 5
    ULLONG = 6
    FLOAT = 7
    DOUBLE = 8
    STRING = 9
    CHAR = 10
    NULL = 11
    BOOL = 12
    NAME = 13
    UNARY = 14
    BINARY = 15


class Eolian_Expression_Mask(IntEnum):
    SINT = 1 << 0
    UINT = 1 << 1
    INT = SINT | UINT
    FLOAT = 1 << 2
    BOOL = 1 << 3
    STRING = 1 << 4
    CHAR = 1 << 5
    NULL = 1 << 6
    SIGNED = SINT | FLOAT
    NUMBER = INT | FLOAT
    ALL = NUMBER | BOOL | STRING | CHAR | NULL


class Eolian_Binary_Operator(IntEnum):
    INVALID = 0
    ADD = 1  # + int, float
    SUB = 2  # - int, float
    MUL = 3  # * int, float
    DIV = 4  # / int, float
    MOD = 5  # % int
    EQ = 6  # == all types
    NQ = 7  # != all types
    GT = 8  # >  int, float
    LT = 9  # <  int, float
    GE = 10  # >= int, float
    LE = 11  # <= int, float
    AND = 12  # && all types
    OR = 13  # || all types
    BAND = 14  # &  int
    BOR = 15  # |  int
    BXOR = 16  # ^  int
    LSH = 17  # << int
    RSH = 18  # >> int


class Eolian_Unary_Operator(IntEnum):
    INVALID = 0
    UNM = 1  # - sint
    UNP = 2  # + sint
    NOT = 3   # ! int, float, bool
    BNOT = 4  # ~ int


class Eolian_Doc_Token_Type(IntEnum):
    UNKNOWN = 0
    TEXT = 1
    REF = 2
    MARK_NOTE = 3
    MARK_WARNING = 4
    MARK_REMARK = 5
    MARK_TODO = 6
    MARKUP_MONOSPACE = 7


#  Internal Classes  ##########################################################

class Iterator(object):
    """ Generic eina iterator wrapper """
    def __init__(self, conv_func, iterator):
        self.next = self.__next__  # py2 compat
        self._conv = conv_func
        self._iter = c_void_p(iterator)
        self._tmp = c_void_p(0)

    def __iter__(self):
        return self

    def __next__(self):
        if not self._iter or not self._iter.value:
            raise StopIteration
        if not lib.eina_iterator_next(self._iter, byref(self._tmp)):
            lib.eina_iterator_free(self._iter)
            raise StopIteration
        return self._conv(self._tmp)

    def free(self):
        lib.eina_iterator_free(self._iter)


class cached_property(object):
    """ Decorator to be used on properties that can be cached """
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__

    def __get__(self, obj, cls):
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class EolianBaseObject(object):
    __instances_cache = {}

    def __new__(cls, c_obj_pointer=None, *args, **kargs):
        # cannot cache without a pointer
        if c_obj_pointer is None:
            return super().__new__(cls)

        # cache based on the c pointer value (assume the eolian db is stable)
        if isinstance(c_obj_pointer, c_void_p):
            key = c_obj_pointer.value
        elif isinstance(c_obj_pointer, int):
            key = c_obj_pointer
        else:
            raise TypeError('Wrong EolianBaseObject constructor pointer!!')

        # get instance from cache or create a new one
        if key in cls.__instances_cache:
            inst = cls.__instances_cache[key]
        else:
            inst = cls.__instances_cache[key] = super().__new__(cls)

        return inst

    def __init__(self, c_obj_pointer):
        if isinstance(c_obj_pointer, c_void_p):
            self._obj = c_void_p(c_obj_pointer.value)
        elif isinstance(c_obj_pointer, int):
            self._obj = c_void_p(c_obj_pointer)
        else:
            raise TypeError('Invalid constructor of type: %s for class: %s' % (
                            type(c_obj_pointer), self.__class__.__name__))

    def __eq__(self, other):
        if isinstance(other, EolianBaseObject):
            return self._obj.value == other._obj.value
        elif isinstance(other, str):
            if hasattr(self, 'name'):
                return self.name == other
            elif hasattr(self, 'short_name'):
                return self.short_name == other
        return False

    def __gt__(self, other):
        if isinstance(other, EolianBaseObject):
            if hasattr(self, 'name'):
                return self.name > other.name
            elif hasattr(self, 'short_name'):
                return self.short_name > other.short_name

    def __hash__(self):
        return self._obj.value

    @property
    def _as_parameter_(self):
        """ Used by ctypes to convert instances when calling C functions """
        return self._obj


#  Main Eolian Unit  ##########################################################

class Eolian_Unit(EolianBaseObject):
    def __repr__(self):
        return "<eolian.Eolian_Unit '{0.file}'>".format(self)

    @property
    def children(self):
        return Iterator(Eolian_Unit, lib.eolian_unit_children_get(self))

    @cached_property
    def file(self):
        return _str_to_py(lib.eolian_unit_file_get(self))

    @cached_property
    def file_path(self):
        return _str_to_py(lib.eolian_unit_file_path_get(self))

    @cached_property
    def state(self):
        c_state = lib.eolian_unit_state_get(self)
        return Eolian_State(c_state) if c_state else None

    @cached_property
    def version(self):
        return lib.eolian_unit_version_get(self)

    @property
    def objects(self):
        return Iterator(Object, lib.eolian_unit_objects_get(self))

    def object_by_name_get(self, name):
        c_obj = lib.eolian_unit_object_by_name_get(self, _str_to_bytes(name))
        return Object(c_obj) if c_obj else None

    @property
    def classes(self):
        return Iterator(Class, lib.eolian_unit_classes_get(self))

    def class_by_name_get(self, class_name):
        c_cls = lib.eolian_unit_class_by_name_get(self, _str_to_bytes(class_name))
        return Class(c_cls) if c_cls else None

    @property
    def constants(self):
        return Iterator(Constant, lib.eolian_unit_constants_get(self))

    def constant_by_name_get(self, name):
        c_var = lib.eolian_unit_constant_by_name_get(self, _str_to_bytes(name))
        return Constant(c_var) if c_var else None

    @property
    def enums(self):
        return Iterator(Typedecl, lib.eolian_unit_enums_get(self))

    def enum_by_name_get(self, name):
        c_tdecl = lib.eolian_unit_enum_by_name_get(self, _str_to_bytes(name))
        return Typedecl(c_tdecl) if c_tdecl else None

    @property
    def structs(self):
        return Iterator(Typedecl, lib.eolian_unit_structs_get(self))

    def struct_by_name_get(self, name):
        c_tdecl = lib.eolian_unit_struct_by_name_get(self, _str_to_bytes(name))
        return Typedecl(c_tdecl) if c_tdecl else None

    @property
    def aliases(self):
        return Iterator(Typedecl, lib.eolian_unit_aliases_get(self))

    def alias_by_name_get(self, name):
        c_tdecl = lib.eolian_unit_alias_by_name_get(self, _str_to_bytes(name))
        return Typedecl(c_tdecl) if c_tdecl else None

    @property
    def all_namespaces(self):
        # TODO find a better way to find namespaces (maybe inside eolian?)
        nspaces = set()
        for obj in self.classes:
            nspaces.add(Namespace(self, obj.namespace))
        for obj in self.aliases:
            nspaces.add(Namespace(self, obj.namespace))
        for obj in self.structs:
            nspaces.add(Namespace(self, obj.namespace))
        for obj in self.enums:
            nspaces.add(Namespace(self, obj.namespace))
        return sorted(nspaces)

    def namespace_get_by_name(self, name):
        return Namespace(self, name)


class Eolian_State(Eolian_Unit):
    def __init__(self, c_state=None):
        if c_state is None:
            c_state = lib.eolian_state_new()  # Eolian_State *
        EolianBaseObject.__init__(self, c_state)

    def __del__(self):
        if not _already_halted:  # do not free after eolian_shutdown
            lib.eolian_state_free(self)

    def __repr__(self):
        return "<eolian.Eolian_State, %d units loaded>" % len(list(self.units))

    def file_parse(self, filename):
        c_unit = lib.eolian_state_file_parse(self, _str_to_bytes(filename))
        return Eolian_Unit(c_unit) if c_unit else None

    def file_path_parse(self, filepath):
        c_unit = lib.eolian_state_file_path_parse(self, _str_to_bytes(filepath))
        return Eolian_Unit(c_unit) if c_unit else None

    @property
    def eo_file_paths(self):
        return Iterator(_str_to_py, lib.eolian_state_eo_file_paths_get(self))

    @property
    def eot_file_paths(self):
        return Iterator(_str_to_py, lib.eolian_state_eot_file_paths_get(self))

    @property
    def eo_files(self):
        return Iterator(_str_to_py, lib.eolian_state_eo_files_get(self))

    @property
    def eot_files(self):
        return Iterator(_str_to_py, lib.eolian_state_eot_files_get(self))

    def directory_add(self, dir_path):
        return bool(lib.eolian_state_directory_add(self, _str_to_bytes(dir_path)))

    def system_directory_add(self):
        return bool(lib.eolian_state_system_directory_add(self))

    def all_eo_files_parse(self):
        return bool(lib.eolian_state_all_eo_files_parse(self))

    def all_eot_files_parse(self):
        return bool(lib.eolian_state_all_eot_files_parse(self))

    def unit_by_file_get(self, file_name):
        c_unit = lib.eolian_state_unit_by_file_get(self, _str_to_bytes(file_name))
        return Eolian_Unit(c_unit) if c_unit else None

    @property
    def units(self):
        return Iterator(Eolian_Unit, lib.eolian_state_units_get(self))

    def objects_by_file_get(self, file_name):
        return Iterator(Object,
                        lib.eolian_state_objects_by_file_get(self, _str_to_bytes(file_name)))

    def class_by_file_get(self, file_name):
        c_cls = lib.eolian_state_class_by_file_get(self, _str_to_bytes(file_name))
        return Class(c_cls) if c_cls else None

    def constants_by_file_get(self, file_name):
        return Iterator(Constant,
                        lib.eolian_state_constants_by_file_get(self, _str_to_bytes(file_name)))

    def aliases_by_file_get(self, file_name):
        return Iterator(Typedecl,
                        lib.eolian_state_aliases_by_file_get(self, _str_to_bytes(file_name)))

    def structs_by_file_get(self, file_name):
        return Iterator(Typedecl,
                        lib.eolian_state_structs_by_file_get(self, _str_to_bytes(file_name)))

    def enums_by_file_get(self, file_name):
        return Iterator(Typedecl,
                        lib.eolian_state_enums_by_file_get(self, _str_to_bytes(file_name)))

    def state_check(self):
        return bool(lib.eolian_state_check(self))


#  Helper functions  ###################################################

def parse_folders(folders):
    """Loads a new database scanning the given folders.

    For example:

    SCAN_FOLDER = os.path.join(eolian.in_tree_src_dir(), 'src', 'lib')
    eolian_db = eolian.parse_folders(SCAN_FOLDER)
    """

    db = Eolian_State()
    if not isinstance(db, Eolian_State):
        raise (RuntimeError('Eolian, failed to create Eolian state'))

    if isinstance(folders, str):
        folders = [folders]

    for folder in folders:
        # eolian source tree scan
        if not db.directory_add(folder):
            raise (RuntimeError('Eolian, failed to scan source dirsectory'))

    # Parse all known eo files
    if not db.all_eot_files_parse():
        raise (RuntimeError('Eolian, failed to parse all EOT files'))

    if not db.all_eo_files_parse():
        raise (RuntimeError('Eolian, failed to parse all EO files'))

    return db

def in_tree_src_dir():
    """Returns the root folder of this script's source tree"""
    script_path = os.path.dirname(os.path.realpath(__file__))
    return os.path.abspath(os.path.join(script_path, '..', '..', '..'))


#  Namespace Utility Class  ###################################################

class Namespace(object):
    def __init__(self, unit, namespace_name):
        self._name = namespace_name
        self._unit = unit

    def __repr__(self):
        return "<eolian.Namespace '{0._name}'>".format(self)

    def __eq__(self, other):
        if isinstance(other, Namespace):
            return self.name == other.name
        if isinstance(other, str):
            return self.name == other
        raise TypeError('Namespace can only compare with Namespace or str')

    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name

    def __hash__(self):
        return hash(self._name)

    @property
    def unit(self):
        return self._unit

    @property
    def name(self):
        return self._name

    @property
    def namespaces(self):
        return self._name.split('.')

    @property
    def sub_namespaces(self):
        base = self._name + '.'
        deep = self._name.count('.') + 1
        return [ns for ns in self._unit.all_namespaces
                if ns.name.startswith(base) and ns.name.count('.') == deep]

    @property
    def classes(self):
        return sorted([c for c in self._unit.classes
                       if c.namespace == self._name])

    @property
    def regulars(self):
        return sorted([c for c in self._unit.classes
                       if c.type == Eolian_Class_Type.REGULAR
                       and c.namespace == self._name])

    @property
    def abstracts(self):
        return sorted([c for c in self._unit.classes
                       if c.type == Eolian_Class_Type.ABSTRACT
                       and c.namespace == self._name])

    @property
    def mixins(self):
        return sorted([c for c in self._unit.classes
                       if c.type == Eolian_Class_Type.MIXIN
                       and c.namespace == self._name])

    @property
    def interfaces(self):
        return sorted([c for c in self._unit.classes
                       if c.type == Eolian_Class_Type.INTERFACE
                       and c.namespace == self._name])

    @property
    def aliases(self):
        return sorted([td for td in self._unit.aliases
                       if td.namespace == self._name])

    @property
    def structs(self):
        return sorted([td for td in self._unit.structs
                       if td.namespace == self._name])

    @property
    def enums(self):
        return sorted([td for td in self._unit.enums
                       if td.namespace == self._name])


#  Eolian Classes  ############################################################

class Object(EolianBaseObject):
    def __new__(cls, c_obj_pointer):
        if cls is Object:
            c_type = lib.eolian_object_type_get(c_obj_pointer)
            cls = _eolian_type_class_mapping[c_type]
        return super().__new__(cls)

    def __repr__(self):
        return "<eolian.Object '{0.name}', {0.type!s}>".format(self)

    @cached_property
    def unit(self):
        c_unit = lib.eolian_object_unit_get(self)
        return Eolian_Unit(c_unit) if c_unit else None

    @cached_property
    def name(self):
        return _str_to_py(lib.eolian_object_name_get(self))

    @cached_property
    def c_name(self):
        return _str_to_py(lib.eolian_object_c_name_get(self))

    @cached_property
    def short_name(self):
        return _str_to_py(lib.eolian_object_short_name_get(self))

    @property
    def namespaces(self):
        return Iterator(_str_to_py, lib.eolian_object_namespaces_get(self))

    @cached_property
    def namespace(self):
        return '.'.join(self.namespaces)

    @cached_property
    def file(self):
        return _str_to_py(lib.eolian_object_file_get(self))

    @cached_property
    def line(self):
        return int(lib.eolian_object_line_get(self))

    @cached_property
    def column(self):
        return int(lib.eolian_object_column_get(self))

    @cached_property
    def is_beta(self):
        return bool(lib.eolian_object_is_beta(self))


class Class(Object):
    def __repr__(self):
        return "<eolian.Class '{0.name}', {0.type!s}>".format(self)

    @cached_property
    def c_macro(self):
        s = lib.eolian_class_c_macro_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def c_get_function_name(self):
        s = lib.eolian_class_c_get_function_name_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def type(self):
        return Eolian_Class_Type(lib.eolian_class_type_get(self))
    
    @cached_property
    def data_type(self):
        return _str_to_py(lib.eolian_class_data_type_get(self))

    @cached_property
    def c_data_type(self):
        s = lib.eolian_class_c_data_type_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def c_prefix(self):
        return _str_to_py(lib.eolian_class_c_prefix_get(self))

    @cached_property
    def event_c_prefix(self):
        return _str_to_py(lib.eolian_class_event_c_prefix_get(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_class_documentation_get(self)
        return Documentation(c_doc) if c_doc else None

    @property
    def constructors(self):
        return Iterator(Constructor, lib.eolian_class_constructors_get(self))

    @property
    def events(self):
        return Iterator(Event, lib.eolian_class_events_get(self))

    def event_by_name_get(self, event_name):
        c_event = lib.eolian_class_event_by_name_get(self,
                                                     _str_to_bytes(event_name))
        return Event(c_event) if c_event else None

    @cached_property
    def parent(self):
        c_class = lib.eolian_class_parent_get(self)
        return Class(c_class) if c_class else None

    @property
    def extensions(self):
        return Iterator(Class, lib.eolian_class_extensions_get(self))

    @cached_property
    def extensions_hierarchy(self):
        visited = set()
        queue = [ext for ext in self.extensions]

        while queue:
            current = queue.pop()

            if current in visited:
                continue

            visited.add(current)

            queue.extend(current.extensions)

        return visited

    @cached_property
    def inherits_full(self):
        li = []

        def do_class_recursive(cls):
            if cls.parent:
                li.append(cls.parent)
                do_class_recursive(cls.parent)
            for other in cls.extensions:
                if other not in li:
                    li.append(other)
                do_class_recursive(other)

        do_class_recursive(self)
        return li

    @cached_property
    def hierarchy(self):
        li = []
        base = self.parent
        while base:
            li.append(base)
            base = base.parent
        return li

    @cached_property
    def ctor_enable(self):
        return bool(lib.eolian_class_ctor_enable_get(self))

    @cached_property
    def dtor_enable(self):
        return bool(lib.eolian_class_dtor_enable_get(self))

    def function_by_name_get(self, function_name,
                             ftype=Eolian_Function_Type.UNRESOLVED):
        f = lib.eolian_class_function_by_name_get(self,
                                                  _str_to_bytes(function_name),
                                                  ftype)
        return Function(f) if f else None

    def functions_get(self, ftype):
        return Iterator(Function, lib.eolian_class_functions_get(self, ftype))

    @property
    def methods(self):
        return self.functions_get(Eolian_Function_Type.METHOD)

    @property
    def properties(self):
        return self.functions_get(Eolian_Function_Type.PROPERTY)

    @property
    def implements(self):
        return Iterator(Implement, lib.eolian_class_implements_get(self))

    @property
    def parts(self):
        return Iterator(Part, lib.eolian_class_parts_get(self))

    @property
    def requires(self):
        return Iterator(Class, lib.eolian_class_requires_get(self))


class Part(Object):
    def __repr__(self):
        return "<eolian.Part '{0.name}'>".format(self)

    @cached_property
    def class_(self):
        return Class(lib.eolian_part_class_get(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_part_documentation_get(self)
        return Documentation(c_doc) if c_doc else None


class Constructor(Object):
    def __repr__(self):
        return "<eolian.Constructor '{0.name}', optional={0.is_optional}>".format(self)

    @cached_property
    def function(self):
        return Function(lib.eolian_constructor_function_get(self))

    @cached_property
    def is_optional(self):
        return bool(lib.eolian_constructor_is_optional(self))

    @cached_property
    def class_(self):
        return Class(lib.eolian_constructor_class_get(self))


class Event(Object):
    def __repr__(self):
        return "<eolian.Event '{0.name}', c_macro='{0.c_macro}'>".format(self)

    @cached_property
    def c_macro(self):
        s = lib.eolian_event_c_macro_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def type(self):
        c_type = lib.eolian_event_type_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def class_(self):
        c_cls = lib.eolian_event_class_get(self)
        return Class(c_cls) if c_cls else None

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_event_documentation_get(self)
        return Documentation(c_doc) if c_doc else None

    @cached_property
    def scope(self):
        return Eolian_Object_Scope(lib.eolian_event_scope_get(self))

    @cached_property
    def is_hot(self):
        return bool(lib.eolian_event_is_hot(self))

    @cached_property
    def is_restart(self):
        return bool(lib.eolian_event_is_restart(self))


class Function(Object):
    def __repr__(self):
        return "<eolian.Function '{0.name}'>".format(self)

    def full_c_name_get(self, ftype):
        s = lib.eolian_function_full_c_name_get(self, ftype)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def full_c_method_name(self):
        return self.full_c_name_get(Eolian_Function_Type.METHOD)

    @cached_property
    def full_c_getter_name(self):
        return self.full_c_name_get(Eolian_Function_Type.PROP_GET)
    
    @cached_property
    def full_c_setter_name(self):
        return self.full_c_name_get(Eolian_Function_Type.PROP_SET)

    @cached_property
    def type(self):
        return Eolian_Function_Type(lib.eolian_function_type_get(self))

    def scope_get(self, ftype):
        return Eolian_Object_Scope(lib.eolian_function_scope_get(self, ftype))

    @cached_property
    def method_scope(self):
        return self.scope_get(Eolian_Function_Type.METHOD)

    @cached_property
    def getter_scope(self):
        return self.scope_get(Eolian_Function_Type.PROP_GET)

    @cached_property
    def setter_scope(self):
        return self.scope_get(Eolian_Function_Type.PROP_SET)

    @cached_property
    def is_static(self):
        return bool(lib.eolian_function_is_static(self))

    @cached_property
    def object_is_const(self):
        return bool(lib.eolian_function_object_is_const(self))

    @cached_property
    def class_(self):
        c_cls = lib.eolian_function_class_get(self)
        return Class(c_cls) if c_cls else None

    def is_constructor(self, klass):
        return bool(lib.eolian_function_is_constructor(self, klass._obj))

    #  @cached_property
    #  def is_function_pointer(self):
        #  return bool(lib.eolian_function_is_function_pointer(self))

    @property
    def parameters(self):
        return Iterator(Function_Parameter,
                        lib.eolian_function_parameters_get(self))

    def values_get(self, ftype):
        # TODO rename in property_values_get
        #  (or implement a proper Property class?)
        return Iterator(Function_Parameter,
                        lib.eolian_property_values_get(self, ftype))

    @property
    def getter_values(self):  # TODO rename ...
        return self.values_get(Eolian_Function_Type.PROP_GET)

    @property
    def setter_values(self):  # TODO rename ...
        return self.values_get(Eolian_Function_Type.PROP_SET)

    def keys_get(self, ftype):
        # TODO rename in property_keys_get
        #  (or implement a proper Property class?)
        return Iterator(Function_Parameter,
                        lib.eolian_property_keys_get(self, ftype))

    @property
    def getter_keys(self):  # TODO rename ...
        return self.keys_get(Eolian_Function_Type.PROP_GET)

    @property
    def setter_keys(self):  # TODO rename ...
        return self.keys_get(Eolian_Function_Type.PROP_SET)

    def return_type_get(self, ftype):
        c_type = lib.eolian_function_return_type_get(self, ftype)
        return Type(c_type) if c_type else None

    def return_c_type_get(self, ftype):
        s = lib.eolian_function_return_c_type_get(self, ftype)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    def return_default_value(self, ftype):
        c_expr = lib.eolian_function_return_default_value_get(self._obj, ftype)
        return Expression(c_expr) if c_expr else None

    def return_documentation(self, ftype):
        c_doc = lib.eolian_function_return_documentation_get(self, ftype)
        return Documentation(c_doc) if c_doc else None

    def return_allow_unused(self, ftype):
        return bool(lib.eolian_function_return_allow_unused(self, ftype))

    def return_is_by_ref(self, ftype):
        return bool(lib.eolian_function_return_is_by_ref(self, ftype))

    def return_is_move(self, ftype):
        return bool(lib.eolian_function_return_is_move(self, ftype))

    @cached_property
    def method_return_type(self):
        return self.return_type_get(Eolian_Function_Type.METHOD)

    @cached_property
    def getter_return_type(self):
        return self.return_type_get(Eolian_Function_Type.PROP_GET)

    @cached_property
    def setter_return_type(self):
        return self.return_type_get(Eolian_Function_Type.PROP_SET)
    
    @cached_property
    def prop_readable(self):
        # TODO: maybe there is a better way to do this...
        ftype = Eolian_Function_Type.PROP_GET
        scope = lib.eolian_function_scope_get(self, ftype)
        return True if scope != Eolian_Object_Scope.UNKNOWN else False

    @cached_property
    def prop_writable(self):
        # TODO: maybe there is a better way to do this...
        ftype = Eolian_Function_Type.PROP_SET
        scope = lib.eolian_function_scope_get(self, ftype)
        return True if scope != Eolian_Object_Scope.UNKNOWN else False

    @cached_property
    def implement(self):
        c_impl = lib.eolian_function_implement_get(self)
        return Implement(c_impl) if c_impl else None


class Function_Parameter(Object):
    def __repr__(self):
        return "<eolian.Function_Parameter '{0.name}', type={0.type}," \
               " optional={0.is_optional}>".format(self)

    @cached_property
    def direction(self):
        return Eolian_Parameter_Direction(lib.eolian_parameter_direction_get(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_parameter_documentation_get(self)
        return Documentation(c_doc) if c_doc else None

    @cached_property
    def is_optional(self):
        return bool(lib.eolian_parameter_is_optional(self))

    @cached_property
    def is_by_ref(self):
        return bool(lib.eolian_parameter_is_by_ref(self))

    @cached_property
    def is_move(self):
        return bool(lib.eolian_parameter_is_move(self))

    @cached_property
    def type(self):
        c_type = lib.eolian_parameter_type_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def default_value(self):
        c_expr = lib.eolian_parameter_default_value_get(self)
        return Expression(c_expr) if c_expr else None

    def c_type_get(self, as_return=False):
        s = lib.eolian_parameter_c_type_get(self, as_return)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret


class Implement(Object):
    def __repr__(self):
        return "<eolian.Implement '{0.name}'>".format(self)

    @cached_property
    def parent(self):
        c_impl = lib.eolian_aux_implement_parent_get(self)
        return Implement(c_impl) if c_impl else None

    @cached_property
    def class_(self):
        c_cls = lib.eolian_implement_class_get(self)
        return Class(c_cls) if c_cls else None

    @cached_property
    def implementing_class(self):
        c_cls = lib.eolian_implement_implementing_class_get(self)
        return Class(c_cls) if c_cls else None

    @cached_property
    def function(self):
        c_func = lib.eolian_implement_function_get(self, None)
        return Function(c_func) if c_func else None

    def documentation_get(self, ftype=Eolian_Function_Type.METHOD):
        # c_doc = lib.eolian_implement_documentation_get(self, ftype)
        c_doc = lib.eolian_aux_implement_documentation_get(self, ftype)
        return Documentation(c_doc) if c_doc else None

    @cached_property
    def documentation_fallback(self):
        c_doc = lib.eolian_aux_implement_documentation_fallback_get(self)
        return Documentation(c_doc) if c_doc else None

    def is_auto(self, ftype=Eolian_Function_Type.METHOD):
        return bool(lib.eolian_implement_is_auto(self, ftype))
    # TODO implement util properties for is_auto

    def is_empty(self, ftype=Eolian_Function_Type.METHOD):
        return bool(lib.eolian_implement_is_empty(self, ftype))
    # TODO implement util properties for is_empty

    def is_pure_virtual(self, ftype=Eolian_Function_Type.METHOD):
        return bool(lib.eolian_implement_is_pure_virtual(self, ftype))
    # TODO implement util properties for is_pure_virtual

    @cached_property
    def is_prop_set(self):
        return bool(lib.eolian_implement_is_prop_set(self))

    @cached_property
    def is_prop_get(self):
        return bool(lib.eolian_implement_is_prop_get(self))

    @property
    def is_property(self):
        return self.is_prop_get or self.is_prop_set

    @property
    def is_method(self):
        return not self.is_property


class Type(Object):
    def __repr__(self):
        # return "<eolian.Type '{0.name}', type: {0.type!s}, c_type: '{0.c_type}'>".format(self)
        return "<eolian.Type '{0.name}', type={0.type!s}>".format(self)

    @cached_property
    def type(self):
        return Eolian_Type_Type(lib.eolian_type_type_get(self))

    @cached_property
    def builtin_type(self):
        return Eolian_Type_Builtin_Type(lib.eolian_type_builtin_type_get(self))

    @cached_property
    def c_type(self):
        s = lib.eolian_type_c_type_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def typedecl(self):
        c_tdecl = lib.eolian_type_typedecl_get(self)
        return Typedecl(c_tdecl) if c_tdecl else None

    @cached_property
    def base_type(self):
        c_type = lib.eolian_type_base_type_get(self)
        return Type(c_type) if c_type else None 

    @cached_property
    def next_type(self):
        c_type = lib.eolian_type_next_type_get(self)
        return Type(c_type) if c_type else None 

    @cached_property
    def aliased_base(self):
        c_type = lib.eolian_type_aliased_base_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def class_(self):
        c_cls = lib.eolian_type_class_get(self)
        return Class(c_cls) if c_cls else None

    @cached_property
    def is_const(self):
        return bool(lib.eolian_type_is_const(self))

    @cached_property
    def is_ptr(self):
        return bool(lib.eolian_type_is_ptr(self))

    @cached_property
    def is_move(self):
        return bool(lib.eolian_type_is_move(self))


class Typedecl(Object):
    def __repr__(self):
        return "<eolian.Typedecl '{0.name}', type={0.type!s}>".format(self)

    @cached_property
    def type(self):
        return Eolian_Typedecl_Type(lib.eolian_typedecl_type_get(self))

    @cached_property
    def c_type(self):
        s = lib.eolian_typedecl_c_type_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def free_func(self):
        return _str_to_py(lib.eolian_typedecl_free_func_get(self))

    @cached_property
    def is_extern(self):
        return bool(lib.eolian_typedecl_is_extern(self))

    @property
    def enum_fields(self):
        return Iterator(Enum_Type_Field,
                        lib.eolian_typedecl_enum_fields_get(self))

    def enum_field_get(self, field):
        c_field = lib.eolian_typedecl_enum_field_get(self, _str_to_bytes(field))
        return Enum_Type_Field(c_field) if c_field else None

    @property
    def struct_fields(self):
        return Iterator(Struct_Type_Field,
                        lib.eolian_typedecl_struct_fields_get(self)) 

    def struct_field_get(self, field):
        c_field = lib.eolian_typedecl_struct_field_get(self, _str_to_bytes(field))
        return Struct_Type_Field(c_field) if c_field else None

    @cached_property
    def base_type(self):
        c_type = lib.eolian_typedecl_base_type_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def aliased_base(self):
        c_type = lib.eolian_typedecl_aliased_base_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_typedecl_documentation_get(self)
        return Documentation(c_doc) if c_doc else None

    @cached_property
    def enum_legacy_prefix(self):
        return _str_to_py(lib.eolian_typedecl_enum_legacy_prefix_get(self))

    @cached_property
    def function_pointer(self):
        c_func = lib.eolian_typedecl_function_pointer_get(self)
        return Function(c_func) if c_func else None


class Enum_Type_Field(Object):
    def __repr__(self):
        return "<eolian.Enum_Type_Field '{0.name}', c_constant='{0.c_constant}'>".format(self)

    @cached_property
    def c_constant(self):
        s = lib.eolian_typedecl_enum_field_c_constant_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def value(self):
        c_expr = lib.eolian_typedecl_enum_field_value_get(self, True)
        return Expression(c_expr) if c_expr else None

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_typedecl_enum_field_documentation_get(self)
        return Documentation(c_doc) if c_doc else None


class Struct_Type_Field(Object):
    def __repr__(self):
        return "<eolian.Struct_Type_Field '{0.name}', type={0.type!s}>".format(self)

    @cached_property
    def type(self):
        c_type = lib.eolian_typedecl_struct_field_type_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def c_type(self):
        s = lib.eolian_typedecl_struct_field_c_type_get(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret

    @cached_property
    def is_by_ref(self):
        return bool(lib.eolian_typedecl_struct_field_is_by_ref(self))

    @cached_property
    def is_move(self):
        return bool(lib.eolian_typedecl_struct_field_is_move(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_typedecl_struct_field_documentation_get(self)
        return Documentation(c_doc) if c_doc else None


class Expression(Object):
    def __repr__(self):
        return "<eolian.Expression type={0.type!s}, serialize='{0.serialize}'>".format(self)

    @cached_property
    def type(self):
        return Eolian_Expression_Type(lib.eolian_expression_type_get(self))

    # TODO: EAPI Eolian_Value eolian_expression_value_get(const Eolian_Expression *expr);

    @cached_property
    def serialize(self):
        s = lib.eolian_expression_serialize(self)
        ret = _str_to_py(s)
        lib.eina_stringshare_del(c_void_p(s))
        return ret
    
    @cached_property
    def binary_operator(self):
        c_op = lib.eolian_expression_binary_operator_get(self)
        return Eolian_Binary_Operator(c_op) if c_op is not None else None

    @cached_property
    def binary_lhs(self):
        c_expr = lib.eolian_expression_binary_lhs_get(self)
        return Expression(c_expr) if c_expr else None

    @cached_property
    def binary_rhs(self):
        c_expr = lib.eolian_expression_binary_rhs_get(self)
        return Expression(c_expr) if c_expr else None

    @cached_property
    def unary_operator(self):
        c_op = lib.eolian_expression_unary_operator_get(self)
        return Eolian_Unary_Operator(c_op) if c_op is not None else None

    @cached_property
    def unary_expression(self):
        c_expr = lib.eolian_expression_unary_expression_get(self)
        return Expression(c_expr) if c_expr is not None else None


class Constant(Object):
    def __repr__(self):
        return "<eolian.Constant '{0.name}', type={0.type!s}, file={0.file}>".format(self)

    @cached_property
    def value(self):
        c_expr = lib.eolian_constant_value_get(self)
        return Expression(c_expr) if c_expr else None

    @cached_property
    def type(self):
        c_type = lib.eolian_constant_type_get(self)
        return Type(c_type) if c_type else None

    @cached_property
    def is_extern(self):
        return bool(lib.eolian_constant_is_extern(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_constant_documentation_get(self)
        return Documentation(c_doc) if c_doc else None


class _Eolian_Doc_Token_Struct(ctypes.Structure):
    _fields_ = [("type", c_int),
                ("text", c_char_p),
                ("text_end", c_char_p)]


class Documentation(Object):
    def __repr__(self):
        t = self.summary if len(self.summary) < 40 else self.summary[:40] + '...'
        return "<eolian.Documentation summary='{}'>".format(t)

    # this is too much for py, just use string.split('\n\n') instead
    # def string_split(self, string):
    #    c_list = lib.eolian_documentation_string_split

    @cached_property
    def summary(self):
        return _str_to_py(lib.eolian_documentation_summary_get(self))

    @cached_property
    def description(self):
        return _str_to_py(lib.eolian_documentation_description_get(self))

    @cached_property
    def since(self):
        return _str_to_py(lib.eolian_documentation_since_get(self))

    @cached_property
    def summary_tokens(self):
        """ return a list of paragraphs, each one is a list of tokens """
        return self._tokenize(self.summary)

    @cached_property
    def description_tokens(self):
        """ return a list of paragraphs, each one is a list of tokens """
        return self._tokenize(self.description)

    def _tokenize(self, full_text):
        paragraphs = []
        if not full_text:
            return paragraphs

        tok = _Eolian_Doc_Token_Struct()
        for paragraph in full_text.split('\n\n'):
            tokens = []
            c_paragraph = _str_to_bytes(paragraph)  # keep c_paragraph alive !

            lib.eolian_doc_token_init(byref(tok))
            next_chunk = lib.eolian_documentation_tokenize(c_paragraph, byref(tok))
            while next_chunk:
                typ = lib.eolian_doc_token_type_get(byref(tok))
                txt = lib.eolian_doc_token_text_get(byref(tok))
                ref_obj = c_void_p(0)
                ref_attr = c_void_p(0)
                ref_type = lib.eolian_doc_token_ref_resolve(byref(tok), self.unit.state, byref(ref_obj), byref(ref_attr))
                tokens.append(Documentation_Token(typ, txt, ref_type, ref_obj, ref_attr))
                lib.free(c_void_p(txt))
                next_chunk = lib.eolian_documentation_tokenize(c_char_p(next_chunk), byref(tok))
            paragraphs.append(tokens)

        return paragraphs


class Documentation_Token(object):
    def __init__(self, c_token_type, c_text, c_ref_type, c_ref_obj, c_ref_attr):
        self._type = Eolian_Doc_Token_Type(c_token_type)
        self._text = _str_to_py(c_text)
        self._ref_type = Eolian_Object_Type(c_ref_type)
        self._ref_obj = Object(c_ref_obj) if c_ref_obj else None
        self._ref_attr = Object(c_ref_attr) if c_ref_attr else None

    def __repr__(self):
        t = self.text if len(self.text) < 40 else self.text[:40] + '...'
        return "<eolian.Doc_Token ({}), text='{}', len={}>".format(
                self.type.name, t, len(self.text))

    @property
    def type(self):
        return self._type

    @property
    def text(self):
        return self._text

    @property
    def ref_type(self):
        return self._ref_type

    @property
    def ref_obj(self):
        return self._ref_obj

    @property
    def ref_attr(self):
        return self._ref_attr


class Error(Object):
    def __repr__(self):
        return "<eolian.Error '{0.name}', message='{0.message}'>".format(self)

    @cached_property
    def message(self):
        return _str_to_py(lib.eolian_error_message_get(self))

    @cached_property
    def documentation(self):
        c_doc = lib.eolian_error_documentation_get(self)
        return Documentation(c_doc) if c_doc else None

    @cached_property
    def is_extern(self):
        return bool(lib.eolian_error_is_extern(self))


#  internal string encode/decode  #############################################

def _str_to_bytes(s):
    return s.encode('utf-8')


def _str_to_py(s):
    if s is None:
        return None    
    if isinstance(s, bytes):
        return s.decode('utf-8')
    if isinstance(s, c_char_p):
        return s.value.decode('utf-8')
    if isinstance(s, c_void_p):
        return cast(s, c_char_p).value.decode('utf-8')
    if isinstance(s, int):
        return cast(s, c_char_p).value.decode('utf-8')
    print('WARNING !!!!!!!!! Unknown type: %s' % type(s))


#  internal Object type -> Class mapping  #####################################

_eolian_type_class_mapping = {
    Eolian_Object_Type.UNKNOWN: Object,
    Eolian_Object_Type.CLASS: Class,
    Eolian_Object_Type.TYPEDECL: Typedecl,
    Eolian_Object_Type.STRUCT_FIELD: Struct_Type_Field,
    Eolian_Object_Type.ENUM_FIELD: Enum_Type_Field,
    Eolian_Object_Type.TYPE: Type,
    Eolian_Object_Type.CONSTANT: Constant,
    Eolian_Object_Type.EXPRESSION: Expression,
    Eolian_Object_Type.FUNCTION: Function,
    Eolian_Object_Type.FUNCTION_PARAMETER: Function_Parameter,
    Eolian_Object_Type.EVENT: Event,
    Eolian_Object_Type.PART: Part,
    Eolian_Object_Type.IMPLEMENT: Implement,
    Eolian_Object_Type.CONSTRUCTOR: Constructor,
    Eolian_Object_Type.DOCUMENTATION: Documentation,
    Eolian_Object_Type.ERROR: Error,
}


#  module init/shutdown  ######################################################

def _cleanup():
    global _already_halted
    lib.eolian_shutdown()
    _already_halted = True


lib.eolian_init()
atexit.register(_cleanup)


#  API coverage statistics  ###################################################

if __name__ == '__main__':
    import sys
    import os
    import re

    # find Eolian.h in source tree
    script_path = os.path.dirname(os.path.realpath(__file__))
    eolian_header = os.path.join(script_path, '..', '..', 'lib', 'eolian', 'Eolian.h')
    eolian_header = os.path.abspath(eolian_header)

    # prepare the two regexp
    flags = re.S | re.M
    FUNC_DEFINED_RE = re.compile(r'^\s*EAPI[\w\s*]*(eolian_\w*)\s*\([\w\s*,]*\);', flags)
    FUNC_USED_RE = re.compile(r'^[^\n#]*lib\.(eolian_[\w]*)\s*\(', flags)

    ENUM_DEFINED_RE = re.compile(r'typedef enum\s*{([^}]*)}\s*(Eolian_[\w]*);', flags)
    ENUM_USED_RE = re.compile(r'^class\s([\w]*)\(IntEnum\):', flags)

    # ignore some know cases (function or enum)
    IGNORED = {
        'eolian_implement_documentation_get',  # we are using eolian_aux_implement_documentation_get
        'eolian_documentation_string_split',  # not needed in python
    }

    # extract functions and enums from Eolian.h
    defined_funcs = []
    defined_enums = []
    with open(eolian_header, 'r') as fh:
        header = fh.read()
    for match in re.finditer(FUNC_DEFINED_RE, header):
        func_name = match.group(1)
        if func_name not in IGNORED:
            defined_funcs.append(func_name)
    for match in re.finditer(ENUM_DEFINED_RE, header):
        enum_name = match.group(2)
        if enum_name not in IGNORED:
            defined_enums.append(enum_name)
    defined_funcs = set(defined_funcs)
    defined_enums = set(defined_enums)

    # extract functions and enums from eolian.py (this file)
    used_funcs = []
    used_enums = []
    with open(__file__, 'r') as fh:
        source = fh.read()
    for match in re.finditer(FUNC_USED_RE, source):
        func_name = match.group(1)
        used_funcs.append(func_name)
    for match in re.finditer(ENUM_USED_RE, source):
        enum_name = match.group(1)
        used_enums.append(enum_name)
    used_funcs = set(used_funcs)
    used_enums = set(used_enums)

    # show general info
    num_def, num_usd = len(defined_funcs), len(used_funcs)
    num_enum_def, num_enum_usd = len(defined_enums), len(used_enums)
    print('Pyolian coverage results')
    print('========================')
    print('Found %d functions defined in Eolian.h (%s)' % (num_def, eolian_header))
    print('Found %d functions used in eolian.py (hopefully not multiline-commented)' % num_usd)
    print('Found %d enums defined in Eolian.h' % num_enum_def)
    print('Found %d enums defined in eolian.py' % num_enum_usd)
    print('Total API coverage %.1f%%' % (num_usd / num_def * 100))
    print()

    # list all missing functions
    missing = defined_funcs - used_funcs
    print('{} Missing functions in eolian.py'.format(len(missing)))
    print('=================================')
    for i, func_name in enumerate(sorted(missing), 1):
        print('{:02d}. {}'.format(i, func_name))
    print()

    # list all missing enums
    missing = defined_enums - used_enums
    if len(missing) > 0:
        print('{} Missing enums in eolian.py'.format(len(missing)))
        print('=================================')
        for i, enum_name in enumerate(sorted(missing), 1):
            print('{:02d}. {}'.format(i, enum_name))
        print()

    # List all functions found in C and Py (--all option)
    if '--all' in sys.argv:
        to_list = (
            (defined_funcs, '{} EAPI functions defined in Eolian.h'),
            (used_funcs, '{} functions called in eolian.py'),
            (defined_enums, '{} Enums defined in Eolian.h'),
            (used_enums, '{} Enums defined in eolian.py'),
        )
        for bag, label in to_list:
            print(label.format(len(bag)))
            print('==============================')
            for i, obj_name in enumerate(sorted(bag), 1):
                print('{:03d}. {}'.format(i, obj_name))
            print()
    else:
        print('Additional arguments')
        print('====================')
        print(' --all   To list all functions and enums found')
        print()



#!/usr/bin/env python3
# encoding: utf-8
"""
Efl documentation generator

Use this script without arguments to generate the full documentation of the Efl
namespace in a folder called 'dokuwiki' (-v to see all generated files)

  --help to see all other options

"""
import os
import sys
import argparse
import atexit


# Use .eo files from the source tree (not the installed ones)
script_path = os.path.dirname(os.path.realpath(__file__))
root_path = os.path.abspath(os.path.join(script_path, '..', '..', '..'))
SCAN_FOLDER = os.path.join(root_path, 'src', 'lib')


# Use pyolian from source (not installed)
pyolian_path = os.path.join(root_path, 'src', 'scripts')
sys.path.insert(0, pyolian_path)
from pyolian import eolian
from pyolian.generator import Template


# parse args
parser = argparse.ArgumentParser(description='Pyolian DocuWiki generator.')
parser.add_argument('--root-path', '-r', metavar='FOLDER', default='dokuwiki',
                    help='where to write files to (root of dokuwiki) '
                         'default to: "./dokuwiki"')
parser.add_argument('--verbose', '-v', action='store_true',
                    help='print a line for each rendered file')
parser.add_argument('--exclude-beta', '-e', action='store_true',
                    help='do not generate docs for class in beta state')
parser.add_argument('--namespace', '-n', metavar='ROOT', default='Efl',
                    help='root namespace of the docs. (default to "Efl")')
_choices = ['start', 'classes', 'enums', 'structs', 'aliases']
parser.add_argument('--step', '-s', metavar='STEP', default=None,
                    choices=_choices,
                    help='A single step to run (default to all), '
                         'valid choices: ' + ', '.join(_choices))
args = parser.parse_args()


# load the whole eolian db (from .eo files in source tree)
eolian_db = eolian.Eolian_State()
if not isinstance(eolian_db, eolian.Eolian_State):
    raise(RuntimeError('Eolian, failed to create Eolian state'))

if not eolian_db.directory_add(SCAN_FOLDER):
    raise(RuntimeError('Eolian, failed to scan source directory'))

if not eolian_db.all_eot_files_parse():
    raise(RuntimeError('Eolian, failed to parse all EOT files'))
    
if not eolian_db.all_eo_files_parse():
    raise(RuntimeError('Eolian, failed to parse all EO files'))


# cleanup the database on exit
def cleanup_db():
    global eolian_db
    del eolian_db


atexit.register(cleanup_db)


# calculate the full path for the txt page of the given object
def page_path_for_object(obj):
    path = ['data', 'pages', 'develop', 'api']
    for ns in obj.namespaces:
        path.append(ns.lower())
    output_filename = obj.short_name.lower() + '.txt'
    return os.path.join(args.root_path, *path, output_filename)


class BetaNamespaceWrapper(eolian.Namespace):
    """ A Namespace wrapper that hide objects marked as beta to the template """
    def __init__(self, eolian_ns):
        super(BetaNamespaceWrapper, self).__init__(eolian_ns.unit, eolian_ns.name)
        self._ns = eolian_ns

    @property
    def sub_namespaces(self):
        return [BetaNamespaceWrapper(ns) for ns in self._ns.sub_namespaces]

    @property
    def classes(self):
        return [c for c in self._ns.classes if not (args.exclude_beta and c.is_beta)]

    @property
    def regulars(self):
        return [c for c in self._ns.regulars if not (args.exclude_beta and c.is_beta)]

    @property
    def abstracts(self):
        return [c for c in self._ns.abstracts if not (args.exclude_beta and c.is_beta)]

    @property
    def mixins(self):
        return [c for c in self._ns.mixins if not (args.exclude_beta and c.is_beta)]

    @property
    def interfaces(self):
        return [c for c in self._ns.interfaces if not (args.exclude_beta and c.is_beta)]

    @property
    def aliases(self):
        return [c for c in self._ns.aliases if not (args.exclude_beta and c.is_beta)]

    @property
    def structs(self):
        return [c for c in self._ns.structs if not (args.exclude_beta and c.is_beta)]

    @property
    def enums(self):
        return [c for c in self._ns.enums if not (args.exclude_beta and c.is_beta)]


# render a (temporary) page for analizying the namespaces hierarchy
t = Template('namespaces.template')
nspaces = [BetaNamespaceWrapper(ns) for ns in eolian_db.all_namespaces
           if ns.name.startswith(args.namespace)]

tot_classes = tot_regulars = tot_abstracts = tot_mixins = tot_ifaces = 0
tot_enums = tot_structs = tot_aliases = 0
for ns in nspaces:
    for cls in ns.classes:
        tot_classes += 1
        if cls.type == eolian.Eolian_Class_Type.REGULAR:
            tot_regulars += 1
        elif cls.type == eolian.Eolian_Class_Type.ABSTRACT:
            tot_abstracts += 1
        elif cls.type == eolian.Eolian_Class_Type.MIXIN:
            tot_mixins += 1
        elif cls.type == eolian.Eolian_Class_Type.INTERFACE:
            tot_ifaces += 1
    tot_enums += len(ns.enums)
    tot_structs += len(ns.structs)
    tot_aliases += len(ns.aliases)


totals = [
    ('Namespaces', len(nspaces)),
    ('ALL Classes', tot_classes),
    ('Regular classes', tot_regulars),
    ('Abstract classes', tot_abstracts),
    ('Mixins', tot_mixins),
    ('Interfaces', tot_ifaces),
    ('Enums', tot_enums),
    ('Structs', tot_structs),
    ('Aliases', tot_aliases),
]

root_ns = BetaNamespaceWrapper(eolian_db.namespace_get_by_name(args.namespace))

output_file = os.path.join(args.root_path, 'data', 'pages', 'develop', 'api', 'namespaces.txt')
t.render(output_file, args.verbose, root_ns=root_ns, totals=totals)


# render the main start.txt page
if args.step in ('start', None):
    t = Template('doc_start.template')

    nspaces = [BetaNamespaceWrapper(ns) for ns in eolian_db.all_namespaces
               if ns.name.startswith(args.namespace)]

    output_file = os.path.join(args.root_path, 'data', 'pages', 'develop', 'api', 'start.txt')
    t.render(output_file, args.verbose, nspaces=nspaces)


# render a page for each Class
if args.step in ('classes', None):
    t = Template('doc_class.template')
    for cls in eolian_db.classes:
        if cls.name.startswith(args.namespace):
            if not (args.exclude_beta and cls.is_beta):
                output_file = page_path_for_object(cls)
                t.render(output_file, args.verbose, cls=cls.name)

# render a page for each Enum
if args.step in ('enums', None):
    t = Template('doc_enum.template')
    for enum in eolian_db.enums:
        if enum.name.startswith(args.namespace):
            if not (args.exclude_beta and enum.is_beta):
                output_file = page_path_for_object(enum)
                t.render(output_file, args.verbose, enum=enum.name)

# render a page for each Struct
if args.step in ('structs', None):
    t = Template('doc_struct.template')
    for struct in eolian_db.structs:
        if struct.name.startswith(args.namespace):
            if not (args.exclude_beta and struct.is_beta):
                output_file = page_path_for_object(struct)
                t.render(output_file, args.verbose, struct=struct.name)

# render a page for each Alias
if args.step in ('aliases', None):
    t = Template('doc_alias.template')
    for alias in eolian_db.aliases:
        if alias.name.startswith(args.namespace):
            if not (args.exclude_beta and alias.is_beta):
                output_file = page_path_for_object(alias)
                t.render(output_file, args.verbose, alias=alias.name)



#!/usr/bin/env python3
# encoding: utf-8
"""
Pyolian test suite
==================

to run all the tests:
> pytest test_eolian.py

"""
import os
import sys

import pytest

from . import eolian


# Use .eo files from the source tree (not the installed ones)
script_path = os.path.dirname(os.path.realpath(__file__))
root_path = os.path.abspath(os.path.join(script_path, '..', '..', '..'))
SCAN_FOLDER = os.path.join(root_path, 'src', 'lib')


# the main Eolian unit state
eolian_db = None


@pytest.fixture(scope='module')
def eolian_db():
    db = eolian.Eolian_State()
    if not isinstance(db, eolian.Eolian_State):
        raise (RuntimeError('Eolian, failed to create Eolian state'))

    # eolian system scan (BROKEN)
    #  if not eolian_db.system_directory_add():
    #  raise(RuntimeError('Eolian, failed to scan system directories'))

    # eolian source tree scan
    if not db.directory_add(SCAN_FOLDER):
        raise (RuntimeError('Eolian, failed to scan source directory'))

    # Parse all known eo files
    if not db.all_eot_files_parse():
        raise (RuntimeError('Eolian, failed to parse all EOT files'))

    if not db.all_eo_files_parse():
        raise (RuntimeError('Eolian, failed to parse all EO files'))

    # tests are executed here
    yield db

    # cleanup eolian
    del db


class TestEolian(object):
    def test_file_format(self):
        v = eolian.file_format_version
        assert isinstance(v, int)
        assert v >= 1


class TestBaseObject(object):
    def test_base_object_equality(self, eolian_db):
        cls1 = eolian_db.class_by_name_get('Efl.Loop_Timer')
        cls2 = eolian_db.class_by_file_get('efl_loop_timer.eo')
        assert isinstance(cls1, eolian.Class)
        assert isinstance(cls2, eolian.Class)
        assert cls1 == cls2
        assert cls1 == 'Efl.Loop_Timer'
        assert cls2 == 'Efl.Loop_Timer'
        assert cls1 != 'another string'
        assert cls1 != 1234
        assert cls1 is not None
        assert cls1 != 0

        enum1 = eolian_db.enum_by_name_get('Efl.Ui.Focus.Direction')
        enum2 = eolian_db.enum_by_name_get('Efl.Ui.Focus.Direction')
        assert isinstance(enum1, eolian.Typedecl)
        assert isinstance(enum2, eolian.Typedecl)
        assert enum1 == enum2
        assert enum1 == 'Efl.Ui.Focus.Direction'
        assert enum2 == 'Efl.Ui.Focus.Direction'
        assert enum1 != 'another string'
        assert enum1 != 1234
        assert enum1 is not None
        assert enum1 != 0

        assert cls1 != enum1


class TestEolianState(object):
    def test_unit_getters(self, eolian_db):
        count = 0
        for unit in eolian_db.units:
            assert isinstance(unit, eolian.Eolian_Unit)
            assert unit.file.endswith(('.eo', '.eot'))
            count += 1
        assert count > 400

        unit = eolian_db.unit_by_file_get('efl_ui_win.eo')
        assert isinstance(unit, eolian.Eolian_Unit)
        assert unit.file == 'efl_ui_win.eo'

    def test_object_getters(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert isinstance(obj, eolian.Object)
        assert type(obj) != eolian.Object
        assert obj.name == 'Efl.Ui.Frame'

        count = 0
        for obj in eolian_db.objects:
            assert isinstance(obj, eolian.Object)
            assert type(obj) != eolian.Object
            count += 1
        assert count > 800

        count = 0
        for obj in eolian_db.objects_by_file_get('efl_loop.eo'):
            assert isinstance(obj, eolian.Object)
            assert type(obj) != eolian.Object
            count += 1
        assert count > 1

    @pytest.mark.skip(reason='Should this return True sooner or later?')
    def test_integrity(self, eolian_db):
        assert eolian_db.state_check() is True


class TestEolianUnit(object):
    def test_unit_get(self, eolian_db):
        unit = eolian_db.unit_by_file_get('efl_ui_win.eo')
        assert isinstance(unit.state, eolian.Eolian_State)
        assert unit.state == eolian_db
        assert isinstance(unit.version, int)
        assert unit.version >= 1

    def test_file_get(self, eolian_db):
        unit = eolian_db.unit_by_file_get('efl_ui_win.eo')
        assert isinstance(unit, eolian.Eolian_Unit)
        assert unit.file == 'efl_ui_win.eo'
        assert unit.file_path.endswith('efl_ui_win.eo')  # full path can change

    @pytest.mark.skip(reason='Skipped until unit/state support is fixed')
    def test_children_listing(self, eolian_db):
        li = list(eolian_db.children)
        assert len(li) > 500
        assert isinstance(li[0], eolian.Eolian_Unit)

    def test_file_listing(self, eolian_db):
        li = list(eolian_db.eo_file_paths)
        assert len(li) > 400
        assert li[0].endswith('.eo')

        li = list(eolian_db.eo_files)
        assert len(li) > 400
        assert li[0].endswith('.eo')

        li = list(eolian_db.eot_file_paths)
        assert len(li) > 10
        assert li[0].endswith('.eot')

        li = list(eolian_db.eot_files)
        assert len(li) > 10
        assert li[0].endswith('.eot')

    def test_object_listing(self, eolian_db):
        unit = eolian_db.unit_by_file_get('efl_ui_win.eo')
        assert unit.object_by_name_get('Efl.Ui.Frame') is None

        obj = unit.object_by_name_get('Efl.Ui.Win')
        assert isinstance(obj, eolian.Object)
        assert type(obj) != eolian.Object
        assert obj.name == 'Efl.Ui.Win'

        count = 0
        for obj in unit.objects:
            assert isinstance(obj, eolian.Object)
            assert type(obj) != eolian.Object
            count += 1
        assert count > 5

    def test_enum_listing(self, eolian_db):
        li = list(eolian_db.enums_by_file_get('efl_ui_win.eo'))
        assert len(li) > 5
        assert isinstance(li[0], eolian.Typedecl)
        assert li[0].type == eolian.Eolian_Typedecl_Type.ENUM

        all_count = 0
        for enum in eolian_db.enums:
            assert isinstance(enum, eolian.Typedecl)
            assert enum.type == eolian.Eolian_Typedecl_Type.ENUM
            all_count += 1
        assert all_count > 50

    def test_struct_listing(self, eolian_db):
        li = list(eolian_db.structs_by_file_get('eina_types.eot'))
        assert len(li) > 10
        assert isinstance(li[0], eolian.Typedecl)
        assert li[0].type in (eolian.Eolian_Typedecl_Type.STRUCT,
                              eolian.Eolian_Typedecl_Type.STRUCT_OPAQUE)

        all_count = 0
        for struct in eolian_db.structs:
            assert isinstance(struct, eolian.Typedecl)
            assert struct.type in (eolian.Eolian_Typedecl_Type.STRUCT,
                                   eolian.Eolian_Typedecl_Type.STRUCT_OPAQUE)
            all_count += 1
        assert all_count > 50

    def test_alias_listing(self, eolian_db):
        li = list(eolian_db.aliases_by_file_get('eina_types.eot'))
        assert len(li) > 2
        assert isinstance(li[0], eolian.Typedecl)

        all_count = 0
        for alias in eolian_db.aliases:
            assert isinstance(alias, eolian.Typedecl)
            assert alias.type in (eolian.Eolian_Typedecl_Type.ALIAS,
                                  eolian.Eolian_Typedecl_Type.FUNCTION_POINTER)  # TODO is this correct ??
            all_count += 1
        assert all_count > 10

    def test_constant_listing(self, eolian_db):
        li = list(eolian_db.constants)
        assert len(li) > 2
        assert isinstance(li[0], eolian.Constant)

        li = list(eolian_db.constants_by_file_get('efl_gfx_stack.eo'))
        assert len(li) > 1
        assert isinstance(li[0], eolian.Constant)

    def test_class_listing(self, eolian_db):
        all_count = 0
        for cls in eolian_db.classes:
            assert isinstance(cls, eolian.Class)
            all_count += 1
        assert all_count > 400


class TestEolianNamespace(object):
    def test_all_namespace(self, eolian_db):
        count = 0
        for ns in eolian_db.all_namespaces:
            assert isinstance(ns, eolian.Namespace)
            count += 1
        assert count > 50

    def test_namespace_vs_class_collision(self, eolian_db):
        colliding_classes = ['Ecore.Audio',
                             'Ecore.Audio.In',
                             'Ecore.Audio.Out',
                             'Ecore.Event.Message',
                             'Ector.Buffer',
                             'Ector.Renderer',
                             'Ector.Renderer.Cairo',
                             'Ector.Renderer.GL',
                             'Ector.Renderer.Gradient',
                             'Ector.Renderer.Software',
                             'Ector.Software.Buffer',
                             'Eio.Sentry',
                             'Eldbus.Model',
                             ]
        for ns in eolian_db.all_namespaces:
            cls = eolian_db.class_by_name_get(ns.name)
            # Some legacy classes are parsed and still make this fail.
            if cls:
                assert cls.name in colliding_classes
            else:
                assert cls is None

    def test_namespace_equality(self, eolian_db):
        ns1 = eolian.Namespace(eolian_db, 'Efl.Io')
        ns2 = eolian.Namespace(eolian_db, 'Efl.Net')
        assert isinstance(ns1, eolian.Namespace)
        assert isinstance(ns2, eolian.Namespace)
        assert ns1 != ns2
        assert ns1 == eolian.Namespace(eolian_db, 'Efl.Io')
        assert ns2 == eolian.Namespace(eolian_db, 'Efl.Net')

    def test_namespace_sorting(self, eolian_db):
        nspaces = eolian_db.all_namespaces
        nspaces.sort(reverse=True)
        assert nspaces[0] > nspaces[-1]
        assert nspaces[1] < nspaces[0]

    def test_namespace_by_name(self, eolian_db):
        ns = eolian.Namespace(eolian_db, 'Efl.Net')
        assert isinstance(ns, eolian.Namespace)
        assert ns.name == 'Efl.Net'
        assert ns.namespaces == ['Efl', 'Net']

        ns = eolian_db.namespace_get_by_name('Efl')
        assert isinstance(ns, eolian.Namespace)
        assert ns.name == 'Efl'

        assert len(ns.classes) > 30
        for cls in ns.classes:
            assert isinstance(cls, eolian.Class)
        assert len(ns.regulars) > 4
        for cls in ns.regulars:
            assert isinstance(cls, eolian.Class)
            assert cls.type == eolian.Eolian_Class_Type.REGULAR
        assert len(ns.mixins) > 0
        for cls in ns.mixins:
            assert isinstance(cls, eolian.Class)
            assert cls.type == eolian.Eolian_Class_Type.MIXIN
        assert len(ns.interfaces) > 15
        for cls in ns.interfaces:
            assert isinstance(cls, eolian.Class)
            assert cls.type == eolian.Eolian_Class_Type.INTERFACE

        assert len(ns.enums) > 1
        for td in ns.enums:
            assert isinstance(td, eolian.Typedecl)
            assert td.type == eolian.Eolian_Typedecl_Type.ENUM
        assert len(ns.aliases) > 0
        for td in ns.aliases:
            assert isinstance(td, eolian.Typedecl)
            # TODO eolian_typedecl_all_aliases_get also return FUNCTION_POINTER
            # is this correct? or an eolian bug ?
            #  assert Equal(td.type, eolian.Eolian_Typedecl_Type.ALIAS)
        assert len(ns.structs) > 2
        for td in ns.structs:
            assert isinstance(td, eolian.Typedecl)
            assert td.type == eolian.Eolian_Typedecl_Type.STRUCT


class TestEolianObject(object):
    def test_object_instance(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert isinstance(obj, eolian.Class)
        assert obj.name == 'Efl.Ui.Frame'

    def test_unit(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert isinstance(obj.unit, eolian.Eolian_Unit)
        assert obj.unit.file == 'efl_ui_frame.eo'

    def test_name(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert obj.name == 'Efl.Ui.Frame'
        assert obj.c_name == 'Efl_Ui_Frame'

    def test_short_name(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert obj.short_name == 'Frame'

    def test_file(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert obj.file == 'efl_ui_frame.eo'

    def test_line(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert isinstance(obj.line, int)
        assert obj.line > 0

    def test_column(self, eolian_db):
        obj = eolian_db.object_by_name_get('Efl.Ui.Frame')
        assert isinstance(obj.column, int)
        assert obj.column > 0


class TestEolianClass(object):
    def test_class(self, eolian_db):
        cls = eolian_db.class_by_file_get('efl_loop_timer.eo')
        assert isinstance(cls, eolian.Class)

        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        assert isinstance(cls, eolian.Class)

        assert cls.name == 'Efl.Loop_Timer'
        assert cls.short_name == 'Loop_Timer'
        assert cls.file == 'efl_loop_timer.eo'
        assert list(cls.namespaces) == ['Efl']
        assert cls.type == eolian.Eolian_Class_Type.REGULAR
        assert isinstance(cls.documentation, eolian.Documentation)
        assert cls.c_prefix is None  # TODO fin a class with a value
        assert cls.event_c_prefix is None  # TODO same as above
        assert cls.data_type is None  # TODO same as above
        assert cls.parent.name == 'Efl.Loop_Consumer'
        assert [c.name for c in cls.extensions] == []
        assert [c.name for c in cls.hierarchy] == ['Efl.Loop_Consumer', 'Efl.Object']
        assert cls.ctor_enable is False
        assert cls.dtor_enable is False
        assert cls.c_get_function_name == 'efl_loop_timer_class_get'
        assert cls.c_macro == 'EFL_LOOP_TIMER_CLASS'
        assert cls.c_data_type == 'Efl_Loop_Timer_Data'
        assert [f.name for f in cls.methods] == ['timer_reset', 'timer_loop_reset', 'timer_delay']
        assert [f.name for f in cls.properties] == ['timer_interval', 'time_pending']
        assert len(list(cls.implements)) > 5
        assert isinstance(list(cls.implements)[0], eolian.Implement)

    def test_mixins_requires(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        assert len(list(cls.requires)) == 0
        cls = eolian_db.class_by_name_get('Efl.File')
        assert len(list(cls.requires)) == 1
        assert list(cls.requires)[0].name == 'Efl.Object'


class TestEolianFunction(object):
    def test_function(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        f = cls.function_by_name_get('timer_delay')
        assert isinstance(f, eolian.Function)
        assert f.name == 'timer_delay'
        assert f.type == eolian.Eolian_Function_Type.METHOD
        assert f.method_scope == eolian.Eolian_Object_Scope.PUBLIC
        assert f.getter_scope == eolian.Eolian_Object_Scope.UNKNOWN  # TODO correct ?
        assert f.setter_scope == eolian.Eolian_Object_Scope.UNKNOWN  # TODO correct ?
        assert f.full_c_method_name == 'efl_loop_timer_delay'
        assert f.full_c_getter_name == 'efl_loop_timer_delay_get'
        assert f.full_c_setter_name == 'efl_loop_timer_delay_set'
        assert f.method_return_type is None  # TODO correct ?
        assert f.setter_return_type is None  # TODO correct ?
        assert f.getter_return_type is None  # TODO correct ?
        assert f.is_static is False
        assert f.is_beta is False
        assert f.is_constructor(cls) is False
        #  # #assert f.is_function_pointer == False  # TODO broken somehow
        assert len(list(f.getter_values)) == 1
        assert len(list(f.getter_values)) == 1
        assert len(list(f.parameters)) == 1
        assert f.return_allow_unused(eolian.Eolian_Function_Type.PROP_GET) is True
        assert f.return_is_by_ref(eolian.Eolian_Function_Type.PROP_GET) is False
        assert f.return_is_move(eolian.Eolian_Function_Type.PROP_GET) is False
        assert f.object_is_const is False
        assert f.class_.name == 'Efl.Loop_Timer'
        assert isinstance(f.implement, eolian.Implement)

    def test_function_parameter(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        f = cls.function_by_name_get('timer_delay')
        p = list(f.parameters)[0]
        assert p.direction == eolian.Eolian_Parameter_Direction.IN
        assert p.name == 'add'
        assert p.default_value is None
        assert p.is_optional is False
        assert p.is_by_ref is False
        assert p.is_move is False
        assert p.type.name == 'double'
        assert p.c_type_get(False) == 'double'
        assert p.c_type_get(True) == 'double'
        assert isinstance(p.documentation, eolian.Documentation)


class TestEolianImplement(object):
    def test_implement(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        f = cls.function_by_name_get('timer_delay')
        im = f.implement
        assert isinstance(im, eolian.Implement)
        assert im.name == 'Efl.Loop_Timer.timer_delay'
        assert isinstance(im.class_, eolian.Class)
        assert isinstance(im.function, eolian.Function)
        assert isinstance(im.implementing_class, eolian.Class)
        assert im.implementing_class.name == 'Efl.Loop_Timer'
        assert isinstance(im.documentation_get(), eolian.Documentation)  # TODO is UNRESOLVED correct ?
        assert im.is_auto() is False
        assert im.is_empty() is False
        assert im.is_pure_virtual() is False
        assert im.is_prop_set is False
        assert im.is_prop_get is False
        assert im.is_property is False
        assert im.is_method is True
        assert im.parent is None

    def test_implement_parent(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Ui.Button')
        im = [im for im in cls.implements if im.short_name == 'content_unset'][0]
        assert isinstance(im, eolian.Implement)
        assert im.name == 'Efl.Content.content_unset'  # TODO is this right??
        assert im.parent.name == 'Efl.Content.content_unset'


class TestEolianEvent(object):
    def test_event(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        assert [e.name for e in cls.events] == ['timer,tick']
        ev = cls.event_by_name_get('timer,tick')
        assert isinstance(ev, eolian.Event)
        assert ev.name == 'timer,tick'
        assert ev.c_macro == 'EFL_LOOP_TIMER_EVENT_TIMER_TICK'
        assert ev.type is None  # TODO is this correct
        assert isinstance(ev.documentation, eolian.Documentation)
        assert ev.scope == eolian.Eolian_Object_Scope.PUBLIC
        assert ev.is_beta is False
        assert ev.is_hot is False
        assert ev.is_restart is False
        assert ev.class_.name == cls.name


class TestEolianPart(object):
    def test_part(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Ui.Popup')
        parts = list(cls.parts)
        assert len(parts) > 0

        part = parts[0]
        assert part.name == 'backwall'
        assert isinstance(part.class_, eolian.Class)
        assert part.class_.name == 'Efl.Ui.Popup_Part_Backwall'
        assert isinstance(part.documentation, eolian.Documentation)


class TestEolianConstructor(object):
    def test_constructor(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Ui.Win')
        ctors = list(cls.constructors)
        assert len(ctors) > 0
        ctor = ctors[0]
        assert isinstance(ctor, eolian.Constructor)
        assert ctor.name == 'Efl.Ui.Win.win_name'
        assert ctor.is_optional is True
        assert isinstance(ctor.class_, eolian.Class)
        assert ctor.class_.name == 'Efl.Ui.Win'
        assert isinstance(ctor.function, eolian.Function)
        assert ctor.function.name == 'win_name'


class TestEolianDocumentation(object):
    def test_documentation(self, eolian_db):
        td = eolian_db.class_by_name_get('Efl.Object')
        doc = td.documentation
        assert isinstance(doc, eolian.Documentation)
        assert isinstance(doc.summary, str)
        assert len(doc.summary) > 10
        assert isinstance(doc.description, str)
        assert len(doc.description) > 20
        assert doc.since == '1.22'


class TestEolianConstant(object):
    def test_constant(self, eolian_db):
        var = eolian_db.constant_by_name_get('Efl.Gfx.Hint_Expand')
        assert isinstance(var, eolian.Constant)
        assert var.name == 'Efl.Gfx.Hint_Expand'
        assert var.short_name == 'Hint_Expand'
        assert var.file == 'efl_gfx_hint.eo'
        assert var.is_extern is False
        assert list(var.namespaces) == ['Efl', 'Gfx']
        assert isinstance(var.documentation, eolian.Documentation)
        assert isinstance(var.type, eolian.Type)
        assert isinstance(var.value, eolian.Expression)
        assert float(var.value.serialize) == +1.0


class TestEolianTypedecl(object):
    def test_typedecl_enum(self, eolian_db):
        td = eolian_db.enum_by_name_get('Efl.Net.Http.Version')
        assert isinstance(td, eolian.Typedecl)
        assert td.name == 'Efl.Net.Http.Version'
        assert td.short_name == 'Version'
        assert td.file == 'efl_net_http_types.eot'
        assert list(td.namespaces) == ['Efl', 'Net', 'Http']
        assert isinstance(td.documentation, eolian.Documentation)
        assert td.base_type is None  # TODO find a better test
        assert td.free_func is None  # TODO find a better test
        assert td.function_pointer is None  # TODO find a better test
        assert td.is_extern is False
        assert len(list(td.enum_fields)) == 3
        assert td.c_type == 'enum Efl_Net_Http_Version { v1_0 = 100, v1_1 = 101, v2_0 = 200 }'

    def test_typedecl_enum_field(self, eolian_db):
        td = eolian_db.enum_by_name_get('Efl.Net.Http.Version')
        field = td.enum_field_get('v1_0')
        assert isinstance(field, eolian.Enum_Type_Field)
        assert field.name == 'v1_0'
        assert field.c_constant == 'EFL_NET_HTTP_VERSION_V1_0'
        assert isinstance(field.documentation, eolian.Documentation)
        assert isinstance(field.value, eolian.Expression)

    def test_typedecl_struct(self, eolian_db):
        td = eolian_db.struct_by_name_get('Efl.Gfx.Color32')
        assert isinstance(td, eolian.Typedecl)
        assert td.name == 'Efl.Gfx.Color32'
        assert td.short_name == 'Color32'
        assert td.file == 'efl_canvas_filter_internal.eo'
        assert list(td.namespaces) == ['Efl', 'Gfx']
        assert isinstance(td.documentation, eolian.Documentation)
        assert td.base_type is None  # TODO find a better test
        assert td.free_func is None  # TODO find a better test
        assert td.function_pointer is None  # TODO find a better test
        assert td.is_extern is False
        assert len(list(td.struct_fields)) == 4
        assert td.c_type == 'struct Efl_Gfx_Color32 { uint8_t r; uint8_t g; uint8_t b; uint8_t a; }'

    def test_typedecl_struct_field(self, eolian_db):
        td = eolian_db.struct_by_name_get('Efl.Gfx.Color32')
        field = td.struct_field_get('b')
        assert isinstance(field, eolian.Struct_Type_Field)
        assert field.name == 'b'
        assert isinstance(field.type, eolian.Type)
        assert isinstance(field.documentation, eolian.Documentation)
        assert field.c_type == "uint8_t"
        assert field.is_by_ref is False
        assert field.is_move is False

    def test_typedecl_alias(self, eolian_db):
        alias = eolian_db.alias_by_name_get('Eina.Error')
        assert isinstance(alias, eolian.Typedecl)
        assert alias.type == eolian.Eolian_Typedecl_Type.ALIAS
        assert alias.name == 'Eina.Error'
        assert alias.short_name == 'Error'
        assert isinstance(alias.aliased_base, eolian.Type)
        assert alias.aliased_base.name == 'int'
        assert alias.c_type == 'typedef int Eina_Error'


class TestEolianType(object):
    def test_type_regular_builtin(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Loop_Timer')
        func = cls.function_by_name_get('timer_delay')
        param = list(func.parameters)[0]
        t = param.type  # type double
        assert isinstance(t, eolian.Type)
        assert t.name == 'double'
        assert t.short_name == 'double'
        assert t.type == eolian.Eolian_Type_Type.REGULAR
        assert t.builtin_type == eolian.Eolian_Type_Builtin_Type.DOUBLE
        assert t.file == 'efl_loop_timer.eo'  # TODO is this correct ?
        assert t.base_type is None  # TODO find a better test
        assert t.next_type is None  # TODO find a better test
        assert t.is_const is False
        assert t.is_ptr is False
        assert t.is_move is False
        assert list(t.namespaces) == []
        assert t.class_ is None
        assert t == t.aliased_base  # TODO find a better test

    def test_type_regular(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Gfx.Entity')
        func = cls.function_by_name_get('geometry')
        param = list(func.setter_values)[0]
        t = param.type  # type Eina.Rect
        assert isinstance(t, eolian.Type)
        assert t.name == 'Eina.Rect'
        assert t.short_name == 'Rect'
        assert t.type == eolian.Eolian_Type_Type.REGULAR
        assert t.builtin_type == eolian.Eolian_Type_Builtin_Type.INVALID
        assert t.file == 'efl_gfx_entity.eo'  # TODO is this correct ?
        assert t.is_const is False
        assert t.is_ptr is False
        assert t.is_move is False
        assert list(t.namespaces) == ['Eina']
        assert t.class_ is None
        assert t == t.aliased_base

        td = t.typedecl
        assert isinstance(td, eolian.Typedecl)
        assert td.name == 'Eina.Rect'

    def test_type_class(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Content')
        func = cls.function_by_name_get('content')
        param = list(func.setter_values)[0]
        t = param.type  # type Efl.Gfx (class interface)
        assert isinstance(t, eolian.Type)
        assert t.name == 'Efl.Gfx.Entity'
        assert t.short_name == 'Entity'
        assert t.type == eolian.Eolian_Type_Type.CLASS
        assert t.builtin_type == eolian.Eolian_Type_Builtin_Type.INVALID
        assert t.file == 'efl_content.eo'  # TODO is this correct ?
        assert t.is_const is False
        assert t.is_ptr is False
        assert t.is_move is False
        assert list(t.namespaces) == ['Efl', 'Gfx']
        assert t == t.aliased_base

        cls = t.class_
        assert isinstance(cls, eolian.Class)
        assert cls.name == 'Efl.Gfx.Entity'


class TestEolianExpression(object):
    def test_expression_simple(self, eolian_db):
        td = eolian_db.enum_by_name_get('Efl.Net.Http.Version')
        field = td.enum_field_get('v1_0')
        exp = field.value
        assert isinstance(exp, eolian.Expression)
        assert exp.type == eolian.Eolian_Expression_Type.INT
        assert exp.serialize == '100'

    def test_expression_unary(self, eolian_db):
        var = eolian_db.constant_by_name_get('Efl.Gfx.Stack_Layer_Min')
        exp = var.value
        assert isinstance(exp, eolian.Expression)
        assert exp.type == eolian.Eolian_Expression_Type.UNARY
        assert float(exp.serialize) == 32768.0  # TODO is this a bug? isn't -1.0 =
        assert exp.unary_operator == eolian.Eolian_Unary_Operator.UNM  # Invalid -> NO
        unary = exp.unary_expression
        assert isinstance(unary, eolian.Expression)
        assert unary.type == eolian.Eolian_Expression_Type.INT
        assert float(exp.serialize) == 32768.0  # Bug too ?

        # TODO test_expression_binary
        #  exp.binary_operator # TODO find a better test (only works for BINARY expr)
        #  exp.binary_lhs # TODO find a better test (only works for BINARY expr)
        #  exp.binary_rhs # TODO find a better test (only works for BINARY expr)

class TestEolianInherits(object):
    def test_inherits_full(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Ui.Widget')
        assert 'Efl.Object' in cls.inherits_full

    def test_extensions_hierarchy(self, eolian_db):
        cls = eolian_db.class_by_name_get('Efl.Ui.Widget')

        # inherited extension
        assert any(x.name == 'Efl.Gfx.Stack' for x in cls.extensions_hierarchy)
        # direct extension
        assert any(x.name == 'Efl.Access.Object' for x in cls.extensions_hierarchy)



#!/usr/bin/env python3

fixture_gen_template = """
static void
_{}_fixture(void)
{{
  _setup_window_and_widget({}, {});
}}
"""

tcase_gen_template = """
static void
{}(TCase *tc)
{{
  tcase_add_checked_fixture(tc, _{}_fixture, NULL);
  tcase_add_checked_fixture(tc, fail_on_errors_setup, fail_on_errors_teardown);
  {}(tc);
}}

"""

file_gen_template = """
#ifdef LIST_DECL
  {}
#else
  {}
#endif
"""

import sys
import json

output_file = sys.argv[-1]
input_files = sys.argv[1:-1]
list_of_tcases = "static const Efl_Test_Case etc[] = {\n"
list_entry = "  {{ \"{}-{}\", {}}},\n"
generated_api = ""

def interface_to_api(interface_name):
  return interface_name+"_behavior_test"

def to_func_name(class_name):
  return class_name.replace('.','_').lower()

def to_class_getter(class_name):
  return class_name.replace('.','_').upper()+'_CLASS'

tests = []

for input_file in input_files:
  with open(input_file, 'r') as content_file:
    content = content_file.read()
    start = content.index('spec-meta-start') + len('spec-meta-start')
    end = content.index('spec-meta-end')
    resulting_json = content[start:end]
    tmp = json.loads(resulting_json)
    if "test-interface" in tmp and "test-widgets" in tmp:
      tests.append(tmp)

widgets = []
widget_hitmap = []

for test in tests:
  interface_test = to_func_name(test["test-interface"])
  if "custom-mapping" in test:
    custom_mapping = test["custom-mapping"]
  else:
    custom_mapping = {}
  for widget_class in test["test-widgets"]:
    if widget_class in custom_mapping:
      widget_test_klass = custom_mapping[widget_class]
    else:
      widget_test_klass = "WIDGET_CLASS"
    combo_name = "_{}_{}".format(to_func_name(interface_test), to_func_name(widget_class));
    combo_klass_test_name = "{}_{}".format(to_func_name(widget_class), to_func_name(widget_test_klass))
    list_of_tcases += list_entry.format(interface_test, to_func_name(widget_class), combo_name)
    generated_api += tcase_gen_template.format(combo_name, combo_klass_test_name, interface_to_api(interface_test))
    if combo_klass_test_name not in widget_hitmap:
      widgets += [{ "widget" : widget_class, "test-class" : widget_test_klass, "combo-name" : combo_klass_test_name}]
      widget_hitmap += [combo_klass_test_name]

for widget in widgets:
  generated_api = fixture_gen_template.format(widget["combo-name"], to_class_getter(widget["widget"]), widget["test-class"]) + generated_api

list_of_tcases += "  { NULL, NULL }\n};"

output = open(output_file, "w")
output.write(file_gen_template.format(list_of_tcases, generated_api))



#!/usr/bin/env python

import re
import os
import sys
import pickle
from optparse import OptionParser

def print_node(a, b, on_tree=False):
    if on_tree:
        rank = "source"
    else:
        rank = "same"

    i_list = ifaces.get(a)
    if i_list:
        for iface in i_list:
            if_instance = a + iface
            if_node_tuple = (if_instance, a) if on_tree else (a, if_instance)

            print '"%s" [label="",shape=circle,width=0.2]' % if_instance
            print '"%s" -> "%s" [label="%s", color=transparent]' \
                % (if_instance, if_instance, iface)
            print '{rank="%s" "%s" -> "%s" [arrowhead="none"];}' \
                % ((rank,) + if_node_tuple)

    i_list = ifaces.get(b)
    if i_list:
        for iface in i_list:
            if_instance = b + iface
            if_node_tuple = (if_instance, b) if on_tree else (b, if_instance)

            print '"%s" [label="",shape=circle,width=0.2]' % if_instance
            print '"%s" -> "%s" [label="%s", color=transparent]' \
                % (if_instance, if_instance, iface)
            print '{rank="%s" "%s" -> "%s" [arrowhead="none"];}' \
                % ((rank,) + if_node_tuple)

    print '"%s" -> "%s"' % (a, b), '[arrowhead="empty"];' \
        if on_tree else '[arrowtail="empty",dir=back];'

def topological_sort(dep_map, value):
    hierarchy = []

    def sort_do(dep_map, value):
        if value is None:
            return

        hierarchy.insert(0, value)
        sort_do(dep_map, dep_map.get(value, None))

    sort_do(dep_map, value)

    return hierarchy

def hierachy_build(files_list):
    cls_map = {"Elm_Widget_Smart_Class": "widget"}

    for path in files_list:
        contents = ''.join(l[:-1] for l in open(path))
        m = re.search(class_re, contents)
        iface_m = re.search(class_iface_re, contents)
        if m is not None:
            items.setdefault(m.group(3), []).append(m.group(1))
            if m.group(2) != m.group(3):
                cls_map[m.group(2)] = m.group(1)
        if iface_m is not None:
            items.setdefault(iface_m.group(3), []).append(iface_m.group(1))
            if iface_m.group(2) != iface_m.group(3):
                cls_map[iface_m.group(2)] = iface_m.group(1)
            #hardcoding scrollable now, for brevity -- it may change in future
            ifaces.setdefault(iface_m.group(1), []).append('scrollable')

    for k, v in items.iteritems():
        clsname = cls_map.get(k, k)
        for c in v:
            hierarchy[c] = clsname

def files_list_build(d):
    files_list = []
    for f in os.listdir(d):
        if f.endswith('.c'):
            files_list.append(d + '/' + f)

    return files_list

#widget name, widget class, parent class
class_re = 'EVAS_SMART_SUBCLASS_NEW.*?,.*?_elm_(\w+).*?,.*?(\w+).*?,.*?(\w+)'

#widget name, widget class, parent class
class_iface_re = \
    'EVAS_SMART_SUBCLASS_IFACE_NEW.*?,.*?_elm_(\w+).*?,.*?(\w+).*?,.*?(\w+)'

usage = "usage: %prog -s <DIRECTORY> -o <OUTPUT_FILE>\n" \
    "       %prog -w <WIDGET_NAME> -i <INPUT_FILE>\n" \
    "       %prog -t -i <INPUT_FILE>\n"
parser = OptionParser(usage=usage)
parser.add_option(
    "-s", "--scan", dest="scan_dir", default=None, type="str",
    help="scan for .h/.c files, at the given path, and build the whole" +
    " widget tree")
parser.add_option(
    "-o", "--output", dest="output_file", default=None, type="str",
    help="path of the output scanning file (widget tree)")

parser.add_option(
    "-w", "--widget", dest="widget", default=None, type="str",
    help="name of an specific widget to generate a hierarchy tree for")
parser.add_option(
    "-i", "--input", dest="input_file", default=None, type="str",
    help="path of the input (widget tree) where to get data from")

parser.add_option(
    "-t", "--tree", action="store_true", dest="tree",
    help="generate the whole hierarchy tree")

opts, args = parser.parse_args()

if (not opts.scan_dir and not opts.widget and not opts.tree) \
      or (opts.scan_dir and opts.widget) \
      or (opts.scan_dir and opts.tree) or \
      (opts.tree and opts.widget):
    sys.exit(parser.print_usage())

if opts.scan_dir and not opts.output_file:
    sys.exit(parser.print_usage())

if opts.widget and not opts.input_file:
    sys.exit(parser.print_usage())

if opts.tree and not opts.input_file:
    sys.exit(parser.print_usage())

items = {}
hierarchy = {}
ifaces = {}

if opts.scan_dir:
    files = files_list_build(opts.scan_dir)
    hierachy_build(files)
    pickle.dump([hierarchy, ifaces], open(opts.output_file, "wb" ))
    sys.exit()

if opts.tree:
    print "digraph elm { node [shape=box];"
    print "rankdir=RL;"

    f = open(opts.input_file)
    if not f:
        sys.exit("Bad input file path")

    hierarchy, ifaces = pickle.load(f)

    for cls, parent in hierarchy.items():
        print_node(cls, parent, True);

    print "}"

if opts.widget:
    print "digraph elm { node [shape=box];"

    f = open(opts.input_file)
    if not f:
        sys.exit("Bad input file path")

    hierarchy, ifaces = pickle.load(f)
    l = topological_sort(hierarchy, opts.widget)

    def pairs(lst):
        for i in range(1, len(lst)):
            yield lst[i-1], lst[i]

    for i1, i2 in pairs(l):
        print_node(i1, i2);

    print "}"



#!/usr/bin/env python3
import sys
import os
from ete3 import Tree, TreeStyle, AttrFace, faces, NodeStyle
import argparse

parser = argparse.ArgumentParser(description='Create a image showing all widgets')
parser.add_argument('sourcedir', metavar='sourcedir', help='the path where to find efls source')

G = parser.parse_args()
sys.path.insert(0, os.path.join(G.sourcedir, 'src', 'scripts', 'pyolian'))

import eolian

SCAN_FOLDER = os.path.join(G.sourcedir, 'src', 'lib')

eolian_db = eolian.Eolian_State()

if not eolian_db.directory_add(SCAN_FOLDER):
  raise(RuntimeError('Eolian, failed to scan source directory'))

# Parse all known eo files
if not eolian_db.all_eot_files_parse():
  raise(RuntimeError('Eolian, failed to parse all EOT files'))

if not eolian_db.all_eo_files_parse():
  raise(RuntimeError('Eolian, failed to parse all EO files'))

widget = eolian_db.class_by_name_get("Efl.Ui.Widget")
assert(widget)

def is_widget(obj, rd = 0):
  if obj == widget:
    return rd
  elif obj.parent != None:
    return is_widget(obj.parent, rd + 1)
  return -1

list_of_widgets = []

for kl in eolian_db.classes:
  widget_depth = is_widget(kl)
  if widget_depth != -1:
    insert = {
      'depth' : widget_depth,
      'name' : kl.name,
      'eolian_obj' : kl,
      'beta' : kl.is_beta,
    }
    list_of_widgets.append(insert)

def layout(node):
    # Add node name to laef nodes
    N = AttrFace("name", fsize=9, fgcolor="black")
    faces.add_face_to_node(N, node, 0)


class Widget_Tree():
  def __init__(self, name, parent, is_beta):
    self.name = name
    self.children = []
    if parent != None:
      self.parent = parent
      parent.children.append(self)
      self.tree_node = parent.tree_node.add_child(None, self.name, 1)
    else:
      self.tree_node = Tree(name = self.name).add_child(None, 'Efl.Ui.Widget', 1)
    nstyle = NodeStyle()
    if is_beta:
      nstyle["fgcolor"] = "red"
    else:
      nstyle["fgcolor"] = "blue"
    self.tree_node.set_style(nstyle)

  def dump(self, prev):
    print(prev + self.name)
    self.children.sort(key = lambda x:x.name)
    for c in self.children:
      c.dump(prev + " ")

sorted_widgets = sorted(list_of_widgets, key=lambda x: x['depth'])
widget_tree = {}

for widget in sorted_widgets:
  if widget['depth'] == 0:
    widget_tree[widget['name']] = Widget_Tree(widget['name'], None, widget['beta'])
  else:
    widget_tree[widget['name']] = Widget_Tree(widget['name'], widget_tree[widget['eolian_obj'].parent.name], widget['beta'])

ts = TreeStyle()
ts.layout_fn = layout
ts.show_leaf_name = False
ts.show_scale = False

print("Created widget-overview.svg in $CWD")

widget_tree['Efl.Ui.Widget'].tree_node.render("widget-overview.svg", units="mm", tree_style=ts)



#!/usr/bin/env python3
# encoding: utf-8
"""

Pyolian template based generator.

This is a really powerfull template-based, output-agnostic eolian generator.
You just need a template file and then you can render it with the
wanted eolian scope (class, namespace, enum, struct, etc...).

For example (from this source folder):

./generator.py test_gen_class.template --cls Efl.Loop.Timer
./generator.py test_gen_namespace.template --ns Efl.Ui

...of course you can pass any other class or namespace to the example above.

The generator is based on the great pyratemp engine (THANKS!), you can find
the full template syntax at: www.simple-is-better.org/template/pyratemp.html

Just keep in mind the syntax is a bit different in this implementation:

    sub_start     = "${"      Was "$!" in original pyratemp (and in docs)
    sub_end       = "}$"      Was "!$" in original pyratemp (and in docs)
    _block_start  = "<!--("
    _block_end    = ")-->"
    comment_start = "#!"
    comment_end   = "!#"


You can also import this module and use the provided Template class if you
are more confortable from within python. To import from outside this directory
you need to hack sys.path in a way that this folder will be available on
PYTHON_PATH, fe:

  pyolian_path = os.path.join(EFL_ROOT_PATH, 'src', 'scripts')
  sys.path.insert(0, pyolian_path)
  from pyolian.generator import Template


"""
import os
import datetime
import atexit

try:
    from . import eolian
    from . import pyratemp
except ImportError:
    import eolian
    import pyratemp


# Use .eo files from the source tree (not the installed ones)
script_path = os.path.dirname(os.path.realpath(__file__))
root_path = os.path.abspath(os.path.join(script_path, '..', '..', '..'))
SCAN_FOLDER = os.path.join(root_path, 'src', 'lib')


# load the whole eolian db
eolian_db = eolian.Eolian_State()
if not isinstance(eolian_db, eolian.Eolian_State):
    raise(RuntimeError('Eolian, failed to create Eolian state'))

if not eolian_db.directory_add(SCAN_FOLDER):
    raise(RuntimeError('Eolian, failed to scan source directory'))

if not eolian_db.all_eot_files_parse():
    raise(RuntimeError('Eolian, failed to parse all EOT files'))
    
if not eolian_db.all_eo_files_parse():
    raise(RuntimeError('Eolian, failed to parse all EO files'))


# cleanup the database on exit
def cleanup_db():
    global eolian_db
    del eolian_db


atexit.register(cleanup_db)


class Template(pyratemp.Template):
    """ Pyolian template based generator.

    You can directly use this class to generate custom outputs based
    on the eolian database and your provided templates.

    Usage is as simple as:
        t = Template(<template_file>)
        t.render(<output_file>, cls=..., ns=..., ...)

    Args:
        filename: Template file to load. (REQUIRED)
        context: User provided context for the template (dict).
    """

    def __init__(self, filename, encoding='utf-8', context=None, escape=None,
                 loader_class=pyratemp.LoaderFile,
                 parser_class=pyratemp.Parser,
                 renderer_class=pyratemp.Renderer,
                 eval_class=pyratemp.EvalPseudoSandbox):

        # Build the global context for the template
        global_ctx = {}
        # user provided context (low pri)
        if context:
            global_ctx.update(context)
        # standard names (not overwritables)
        global_ctx.update({
            # Template info
            'date': datetime.datetime.now(),
            'template_file': os.path.basename(filename),
            # Eolian info
            #  'eolian_version': eolian.__version__,
            #  'eolian_version_info': eolian.__version_info__,
            # Eolian Classes
            'Object': eolian.Object,
            'Class': eolian.Class,
            'Part': eolian.Part,
            'Constructor': eolian.Constructor,
            'Event': eolian.Event,
            'Function': eolian.Function,
            'Function_Parameter': eolian.Function_Parameter,
            'Implement': eolian.Implement,
            'Type': eolian.Type,
            'Typedecl': eolian.Typedecl,
            'Enum_Type_Field': eolian.Enum_Type_Field,
            'Struct_Type_Field': eolian.Struct_Type_Field,
            'Expression': eolian.Expression,
            'Constant': eolian.Constant,
            'Documentation': eolian.Documentation,
            'Documentation_Token': eolian.Documentation_Token,
            'Error': eolian.Error,
            # Eolian Enums
            'Eolian_Object_Type': eolian.Eolian_Object_Type,
            'Eolian_Function_Type': eolian.Eolian_Function_Type,
            'Eolian_Parameter_Direction': eolian.Eolian_Parameter_Direction,
            'Eolian_Class_Type': eolian.Eolian_Class_Type,
            'Eolian_Object_Scope': eolian.Eolian_Object_Scope,
            'Eolian_Typedecl_Type': eolian.Eolian_Typedecl_Type,
            'Eolian_Type_Type': eolian.Eolian_Type_Type,
            'Eolian_Type_Builtin_Type': eolian.Eolian_Type_Builtin_Type,
            # 'Eolian_C_Type_Type': eolian.Eolian_C_Type_Type,
            'Eolian_Expression_Type': eolian.Eolian_Expression_Type,
            'Eolian_Expression_Mask': eolian.Eolian_Expression_Mask,
            'Eolian_Binary_Operator': eolian.Eolian_Binary_Operator,
            'Eolian_Unary_Operator': eolian.Eolian_Unary_Operator,
            'Eolian_Doc_Token_Type': eolian.Eolian_Doc_Token_Type,
        })

        # Call the parent __init__ func
        self.template_filename = filename
        pyratemp.Template.__init__(self, filename=filename, encoding=encoding,
                                   data=global_ctx, escape=escape,
                                   loader_class=loader_class,
                                   parser_class=parser_class,
                                   renderer_class=renderer_class,
                                   eval_class=eval_class)

    def render(self, filename=None, verbose=True, cls=None, ns=None,
               struct=None, enum=None, alias=None, **kargs):
        # Build the context for the template
        ctx = {}
        if kargs:
            ctx.update(kargs)
        if cls:
            ctx['cls'] = eolian_db.class_by_name_get(cls)
        if ns:
            ctx['namespace'] = eolian_db.namespace_get_by_name(ns)
        if struct:
            ctx['struct'] = eolian_db.struct_by_name_get(struct)
        if enum:
            ctx['enum'] = eolian_db.enum_by_name_get(enum)
        if alias:
            ctx['alias'] = eolian_db.alias_by_name_get(alias)

        if verbose and filename:
            print("rendering: {} => {}".format(
                  self.template_filename, filename))

        # render with the augmented context
        output = self(**ctx)

        if filename is not None:
            # create directory tree if needed
            folder = os.path.dirname(filename)
            if folder and not os.path.isdir(folder):
                os.makedirs(folder)
            # write to file
            with open(filename, "w") as f:
                f.write(output)
        else:
            # or print to stdout
            print(output)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Pyolian template based generator.')
    parser.add_argument('template',
                        help='The template file to use. (REQUIRED)')
    parser.add_argument('--output', '-o', metavar='FILE', default=None,
                        help='Where to write the rendered output. '
                             'If not given will print to stdout.')
    parser.add_argument('--cls', metavar='CLASS_NAME', default=None,
                        help='The full name of the class to render, ex: Efl.Loop.Timer')
    parser.add_argument('--ns', metavar='NAMESPACE', default=None,
                        help='The namespace to render, ex: Efl.Loop')
    parser.add_argument('--struct', metavar='STRUCT_NAME', default=None,
                        help='The name of the struct to render, ex: Efl.Loop.Arguments')
    parser.add_argument('--enum', metavar='ENUM_NAME', default=None,
                        help='The name of the enum to render, ex: Efl.Loop.Handler.Flags')
    parser.add_argument('--alias', metavar='ALIAS_NAME', default=None,
                        help='The name of the alias to render, ex: Efl.Font.Size')
    args = parser.parse_args()

    t = Template(args.template)
    t.render(args.output, cls=args.cls, ns=args.ns,
             struct=args.struct, enum=args.enum, alias=args.alias)
